--=============================================================================
-- Description: IBIS macro library implemented in VHDL-A(MS)
-- Created: January 2006 by Arpad Muranyi (Intel Corporation)
--
-- Revision history:
--
-- 2006/01/19  VHDL-A(MS) rev0.0 completed
-- 2006/02/01  - Added file reader routine to read parameters from a file
--             - Separated functions into package file
--             - Changed the type of "i12" and "i21" from real to current to
--               eliminate some warning messages
-- 2006/02/06  - Fixed a bug in the IBIS buffer models "FindCommonLength"
--               function was given the wrong vectors
--             - Changed the algorithm of "DIV" controlled sources to handle
--               div/0 situation more gently
--             - Changed the type of "kpd" and "kpu" from real to current to
--               eliminate all warning messages
--             - Added three more data files to test the IBIS buffer models.
--               These files were extracted with the new Perl script from a
--               real IBIS file, and contain full length IV and Vt tables.
-- 2006/02/27  - Fixed a bug in the "FindCommonLength" and "CommonTime"
--               functions of IBIS buffer models.
--             - Added IV and Vt curve scaling capability to IBIS buffers
--             - Corrected kC_comp_** initial values and analog equations 
--               for missing C_comp components in Open_*** models
--             - Removed four unused signals (variables) from IBIS buffers
--
--=============================================================================
-- List of building blocks contained in this library:
--
-- IBIS_R             (p, n);
-- IBIS_VCR           (p, n, ps, ns);
-- IBIS_CCR           (p, n, ps, ns);
--
-- IBIS_C             (p, n);
-- IBIS_VCC           (p, n, ps, ns);
-- IBIS_CCC           (p, n, ps, ns);
--
-- IBIS_L             (p, n);
-- IBIS_VCL           (p, n, ps, ns);
-- IBIS_CCL           (p, n, ps, ns);
-- IBIS_K             (p1, n1, p2, n2);
--
-- IBIS_V             (p, n);
-- IBIS_VCVS          (p, n, ps, ns);
-- IBIS_CCVS          (p, n, ps, ns);
-- IBIS_VCVS_DELAY    (p, n, ps, ns);
-- IBIS_CCVS_DELAY    (p, n, ps, ns);
-- IBIS_VCVS_MIN      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_CCVS_MIN      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_VCVS_MAX      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_CCVS_MAX      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_VCVS_ABS      (p, n, ps, ns);
-- IBIS_CCVS_ABS      (p, n, ps, ns);
-- IBIS_VCVS_SUM      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_CCVS_SUM      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_VCVS_MULT     (p, n, ps1, ns1, ps2, ns2);
-- IBIS_CCVS_MULT     (p, n, ps1, ns1, ps2, ns2);
-- IBIS_VCVS_DIV      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_CCVS_DIV      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_VCVS_PWL      (p, n, ps, ns);
-- IBIS_CCVS_PWL      (p, n, ps, ns);
-- IBIS_TCVS_PWL      (p, n);
-- IBIS_VECVS_PWL     (p, n, ps, ns);
-- IBIS_CECVS_PWL     (p, n, ps, ns);
--
-- IBIS_I             (p, n);
-- IBIS_VCCS          (p, n, ps, ns);
-- IBIS_CCCS          (p, n, ps, ns);
-- IBIS_VCCS_DELAY    (p, n, ps, ns);
-- IBIS_CCCS_DELAY    (p, n, ps, ns);
-- IBIS_VCCS_MIN      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_CCCS_MIN      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_VCCS_MAX      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_CCCS_MAX      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_VCCS_ABS      (p, n, ps, ns);
-- IBIS_CCCS_ABS      (p, n, ps, ns);
-- IBIS_VCCS_SUM      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_CCCS_SUM      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_VCCS_MULT     (p, n, ps1, ns1, ps2, ns2);
-- IBIS_CCCS_MULT     (p, n, ps1, ns1, ps2, ns2);
-- IBIS_VCCS_DIV      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_CCCS_DIV      (p, n, ps1, ns1, ps2, ns2);
-- IBIS_VCCS_PWL      (p, n, ps, ns);
-- IBIS_CCCS_PWL      (p, n, ps, ns);
-- IBIS_TCCS_PWL      (p, n);
-- IBIS_VECCS_PWL     (p, n, ps, ns);
-- IBIS_CECCS_PWL     (p, n, ps, ns);
--
-- IBIS_T             (n1, ref1, n2, ref2);
--
-- IBIS_INPUT         (PC_ref, GC_ref, Input, Rcv_D);
-- IBIS_OUTPUT        (PU_ref, PD_ref, Output, In_D, PC_ref, GC_ref);
-- IBIS_IO            (PU_ref, PD_ref, IO, In_D, En_D, Rcv_D, PC_ref, GC_ref);
-- IBIS_3STATE        (PU_ref, PD_ref, IO, In_D, En_D, PC_ref, GC_ref);
-- IBIS_OPENSINK      (PU_ref, PD_ref, IO, In_D, PC_ref, GC_ref);
-- IBIS_IO_OPENSINK   (PU_ref, PD_ref, IO, In_D, En_D, Rcv_D, PC_ref, GC_ref);
-- IBIS_OPENSOURCE    (PU_ref, PD_ref, IO, In_D, PC_ref, GC_ref);
-- IBIS_IO_OPENSOURCE (PU_ref, PD_ref, IO, In_D, En_D, Rcv_D, PC_ref, GC_ref);
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_R is

    generic (Rval  : real := 1.0;
             Scale : real := 1.0);

    port (terminal p, n : electrical);

  end entity IBIS_R;
  --===========================================================================
  architecture ideal of IBIS_R is

  quantity Vout across Iout through p to n;

  begin

    Vout == Scale * Rval * Iout;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCR is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCR;
  --===========================================================================
  architecture ideal of IBIS_VCR is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Vout == Scale * Vin * Iout;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCR is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCR;
  --===========================================================================
  architecture ideal of IBIS_CCR is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Vout == Scale * Iin * Iout;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_C is

    generic (Cval  : real := 1.0;
             V0    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n : electrical);

  end entity IBIS_C;
  --===========================================================================
  architecture ideal of IBIS_C is

  quantity Vout across Iout through p to n;

  begin

    if domain = quiescent_domain use
      Vout == V0;
    else
      Iout == Scale * Cval * Vout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCC is

    generic (V0    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCC;
  --===========================================================================
  architecture ideal of IBIS_VCC is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    if domain = quiescent_domain use
      Vout == V0;
    else
      Iout == Vout * Scale * Vin'dot + Scale * Vin * Vout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCC is

    generic (V0    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCC;
  --===========================================================================
  architecture ideal of IBIS_CCC is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    if domain = quiescent_domain use
      Vout == V0;
    else
      Iout == Vout * Scale * Iin'dot + Scale * Iin * Vout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_L is

    generic (Lval  : real := 1.0;
             I0    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n : electrical);

  end entity IBIS_L;
  --===========================================================================
  architecture ideal of IBIS_L is

  quantity Vout across Iout through p to n;

  begin

    if domain = quiescent_domain use
      Iout == I0;
    else
      Vout == Scale * Lval * Iout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCL is

    generic (I0    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCL;
  --===========================================================================
  architecture ideal of IBIS_VCL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    if domain = quiescent_domain use
      Iout == I0;
    else
      Vout == Iout * Scale * Vin'dot + Scale * Vin * Iout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCL is

    generic (I0    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCL;
  --===========================================================================
  architecture ideal of IBIS_CCL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    if domain = quiescent_domain use
      Iout == I0;
    else
      Vout == Iout * Scale * Iin'dot + Scale * Iin * Iout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_K is

    generic (Lval_1 : real := 1.0;
             Lval_2 : real := 1.0;
             Kval   : real := 0.0;
             I0_1   : real := 0.0;
             I0_2   : real := 0.0;
             Scale  : real := 1.0);

    port (terminal p1, n1, p2, n2 : electrical);

  end entity IBIS_K;
  --===========================================================================
  architecture ideal of IBIS_K is

  quantity Vout1 across Iout1 through p1 to n1;
  quantity Vout2 across Iout2 through p2 to n2;
  constant M : real := Scale * Kval * sqrt(Lval_1 * Lval_2);

  begin

    if domain = quiescent_domain use
      Iout1 == I0_1;
      Iout2 == I0_2;
    else
      Vout1 == Scale * Lval_1 * Iout1'dot + M * Iout2'dot;
      Vout2 == Scale * Lval_2 * Iout2'dot + M * Iout1'dot;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_V is

    generic (Vdc   : real := 1.0;
             Scale : real := 1.0);

    port (terminal p, n : electrical);

  end entity IBIS_V;
  --===========================================================================
  architecture ideal of IBIS_V is

  quantity Vout across Iout through p to n;

  begin

    Vout == Scale * Vdc;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCVS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCVS;
  --===========================================================================
  architecture ideal of IBIS_VCVS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Vout == Scale * Vin;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCVS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCVS;
  --===========================================================================
  architecture ideal of IBIS_CCVS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Vout == Scale * Iin;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCVS_DELAY is

    generic (TD    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCVS_DELAY;
  --===========================================================================
  architecture ideal of IBIS_VCVS_DELAY is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Vout == Scale * Vin'delayed(TD);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCVS_DELAY is

    generic (TD    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCVS_DELAY;
  --===========================================================================
  architecture ideal of IBIS_CCVS_DELAY is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Vout == Scale * Iin'delayed(TD);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_VCVS_MIN is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCVS_MIN;
  --===========================================================================
  architecture ideal of IBIS_VCVS_MIN is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Vout == Scale * realmin(Vin1, Vin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_CCVS_MIN is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCVS_MIN;
  --===========================================================================
  architecture ideal of IBIS_CCVS_MIN is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Vout == Scale * realmin(Iin1, Iin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_VCVS_MAX is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCVS_MAX;
  --===========================================================================
  architecture ideal of IBIS_VCVS_MAX is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Vout == Scale * realmax(Vin1, Vin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_CCVS_MAX is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCVS_MAX;
  --===========================================================================
  architecture ideal of IBIS_CCVS_MAX is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Vout == Scale * realmax(Iin1, Iin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCVS_ABS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCVS_ABS;
  --===========================================================================
  architecture ideal of IBIS_VCVS_ABS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Vout == Scale * abs(Vin);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCVS_ABS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCVS_ABS;
  --===========================================================================
  architecture ideal of IBIS_CCVS_ABS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Vout == Scale * abs(Iin);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCVS_SUM is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCVS_SUM;
  --===========================================================================
  architecture ideal of IBIS_VCVS_SUM is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Vout == Scale * (Vin1 + Vin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCVS_SUM is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCVS_SUM;
  --===========================================================================
  architecture ideal of IBIS_CCVS_SUM is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Vout == Scale * (Iin1 + Iin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCVS_MULT is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCVS_MULT;
  --===========================================================================
  architecture ideal of IBIS_VCVS_MULT is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Vout == Scale * Vin1 * Vin2;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCVS_MULT is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCVS_MULT;
  --===========================================================================
  architecture ideal of IBIS_CCVS_MULT is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Vout == Scale * Iin1 * Iin2;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCVS_DIV is

    generic (Scale     : real := 1.0;
             ZeroLimit : real := 1.0e-3);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCVS_DIV;
  --===========================================================================
  architecture ideal of IBIS_VCVS_DIV is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    if (Vin2 < -ZeroLimit) or (Vin2 > ZeroLimit) use
      Vout == Scale * Vin1 / Vin2;
    else
      Vout == Scale * Vin1 * Vin2 / ZeroLimit**2;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCVS_DIV is

    generic (Scale     : real := 1.0;
             ZeroLimit : real := 1.0e-3);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCVS_DIV;
  --===========================================================================
  architecture ideal of IBIS_CCVS_DIV is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    if (Iin2 < -ZeroLimit) or (Iin2 > ZeroLimit) use
      Vout == Scale * Iin1 / Iin2;
    else
      Vout == Scale * Iin1 * Iin2 / ZeroLimit**2;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_VCVS_PWL is

    generic (Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             X_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
             Y_data : real_vector := (-0.10,  0.00,  0.10,  0.10);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCVS_PWL;
  --===========================================================================
  architecture ideal of IBIS_VCVS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;
  --===========================================================================
  constant  X_vec : real_vector := FileRead(DataFile, "X_data", X_data, Max_length);
  constant  Y_vec : real_vector := FileRead(DataFile, "Y_data", Y_data, Max_length);
--=============================================================================
  begin

    Vout == Scale * Lookup(Vin, X_vec, Y_vec, "SE");

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_CCVS_PWL is

    generic (Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             X_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
             Y_data : real_vector := (-0.10,  0.00,  0.10,  0.10);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCVS_PWL;
  --===========================================================================
  architecture ideal of IBIS_CCVS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;
  --===========================================================================
  constant  X_vec : real_vector := FileRead(DataFile, "X_data", X_data, Max_length);
  constant  Y_vec : real_vector := FileRead(DataFile, "Y_data", Y_data, Max_length);
--=============================================================================
  begin

    Vin  == 0.0;
    Vout == Scale * Lookup(Iin, X_vec, Y_vec, "SE");

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_TCVS_PWL is

    generic (Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             X_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             Y_data : real_vector := (0.00,     0.10,     0.90,     1.00);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n : electrical);

  end entity IBIS_TCVS_PWL;
  --===========================================================================
  architecture ideal of IBIS_TCVS_PWL is

  quantity Vout across Iout through p to n;
  --===========================================================================
  constant  X_vec : real_vector := FileRead(DataFile, "X_data", X_data, Max_length);
  constant  Y_vec : real_vector := FileRead(DataFile, "Y_data", Y_data, Max_length);
--=============================================================================
  begin

    Vout == Scale * Lookup(now, X_vec, Y_vec, "HE");

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_VECVS_PWL is

    generic (Vth_R : real := 0.0;
             Vth_F : real := 0.0;
             Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             XR_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YR_data : real_vector := (0.00,     0.10,     0.90,     1.00);
             XF_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YF_data : real_vector := (1.00,     0.90,     0.10,     0.00);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VECVS_PWL;
  --===========================================================================
  architecture ideal of IBIS_VECVS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;
  signal   T_event    : real    := 0.0;
  signal   EventState : integer := 0;
  --===========================================================================
  constant  XR_vec : real_vector := FileRead(DataFile, "XR_data", XR_data, Max_length);
  constant  YR_vec : real_vector := FileRead(DataFile, "YR_data", YR_data, Max_length);
  constant  XF_vec : real_vector := FileRead(DataFile, "XF_data", XF_data, Max_length);
  constant  YF_vec : real_vector := FileRead(DataFile, "YF_data", YF_data, Max_length);
--=============================================================================
  begin

    ---------------------------------------------------------------------------
    Threshold_crossing : process is
    begin
      if domain = quiescent_domain then
        wait until domain /= quiescent_domain;
        if Vin > Vth_R then            -- high state
          EventState <= 1;
        elsif Vin < Vth_F then         -- low state
          EventState <= -1;
        end if;
      else
        wait on Vin'above(Vth_R), Vin'above(Vth_F);
        if Vin > Vth_R then            -- rising edge
          EventState <= 2;
          T_event    <= now;
        elsif Vin < Vth_F then         -- falling edge
          EventState <= -2;
          T_event    <= now;
        end if;
      end if;
    end process Threshold_crossing;
    ---------------------------------------------------------------------------

    case EventState use
      when -2 =>        -- Executes after falling edge events
        Vout == Scale * Lookup(now-T_event, XF_vec, YF_vec, "HE");
      when -1 =>        -- Executes between t=0 and first threshold crossing
        Vout == Scale * Lookup(0.0, XR_vec, YR_vec, "HE");    -- when input is LOW
      when 1 =>         -- Executes between t=0 and first threshold crossing
        Vout == Scale * Lookup(0.0, XF_vec, YF_vec, "HE");    -- when input is HIGH
      when 2 =>         -- Executes after rising edge events
        Vout == Scale * Lookup(now-T_event, XR_vec, YR_vec, "HE");
      when others =>    -- Executes during DC operating point analysis
        if Vin > Vth_R use
          Vout == Scale * Lookup(0.0, XF_vec, YF_vec, "HE");
        elsif Vin < Vth_F use
          Vout == Scale * Lookup(0.0, XR_vec, YR_vec, "HE");
        else
          Vout == Scale * (Lookup(0.0, XR_vec, YR_vec, "HE") + Lookup(0.0, XF_vec, YF_vec, "HE")) / 2.0;
        end use;
    end case;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_CECVS_PWL is

    generic (Ith_R : real := 0.0;
             Ith_F : real := 0.0;
             Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             XR_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YR_data : real_vector := (0.00,     0.10,     0.90,     1.00);
             XF_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YF_data : real_vector := (1.00,     0.90,     0.10,     0.00);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CECVS_PWL;
  --===========================================================================
  architecture ideal of IBIS_CECVS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;
  signal   T_event    : real    := 0.0;
  signal   EventState : integer := 0;
  --===========================================================================
  constant  XR_vec : real_vector := FileRead(DataFile, "XR_data", XR_data, Max_length);
  constant  YR_vec : real_vector := FileRead(DataFile, "YR_data", YR_data, Max_length);
  constant  XF_vec : real_vector := FileRead(DataFile, "XF_data", XF_data, Max_length);
  constant  YF_vec : real_vector := FileRead(DataFile, "YF_data", YF_data, Max_length);
--=============================================================================
  begin

    ---------------------------------------------------------------------------
    Threshold_crossing : process is
    begin
      if domain = quiescent_domain then
        wait until domain /= quiescent_domain;
        if Iin > Ith_R then            -- high state
          EventState <= 1;
        elsif Iin < Ith_F then         -- low state
          EventState <= -1;
        end if;
      else
        wait on Iin'above(Ith_R), Iin'above(Ith_F);
        if Iin > Ith_R then            -- rising edge
          EventState <= 2;
          T_event    <= now;
        elsif Iin < Ith_F then         -- falling edge
          EventState <= -2;
          T_event    <= now;
        end if;
      end if;
    end process Threshold_crossing;
    ---------------------------------------------------------------------------

    Vin == 0.0;
    case EventState use
      when -2 =>        -- Executes after falling edge events
        Vout == Scale * Lookup(now-T_event, XF_vec, YF_vec, "HE");
      when -1 =>        -- Executes between t=0 and first threshold crossing
        Vout == Scale * Lookup(0.0, XR_vec, YR_vec, "HE");    -- when input is LOW
      when 1 =>         -- Executes between t=0 and first threshold crossing
        Vout == Scale * Lookup(0.0, XF_vec, YF_vec, "HE");    -- when input is HIGH
      when 2 =>         -- Executes after rising edge events
        Vout == Scale * Lookup(now-T_event, XR_vec, YR_vec, "HE");
      when others =>    -- Executes during DC operating point analysis
        if Iin > Ith_R use
          Vout == Scale * Lookup(0.0, XF_vec, YF_vec, "HE");
        elsif Iin < Ith_F use
          Vout == Scale * Lookup(0.0, XR_vec, YR_vec, "HE");
        else
          Vout == Scale * (Lookup(0.0, XR_vec, YR_vec, "HE") + Lookup(0.0, XF_vec, YF_vec, "HE")) / 2.0;
        end use;
    end case;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_I is

    generic (Idc   : real := 1.0;
             Scale : real := 1.0);

    port (terminal p, n : electrical);

  end entity IBIS_I;
  --===========================================================================
  architecture ideal of IBIS_I is

  quantity Vout across Iout through p to n;

  begin

    Iout == Scale * Idc;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCCS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCCS;
  --===========================================================================
  architecture ideal of IBIS_VCCS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Iout == Scale * Vin;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCCS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCCS;
  --===========================================================================
  architecture ideal of IBIS_CCCS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Iout == Scale * Iin;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCCS_DELAY is

    generic (TD    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCCS_DELAY;
  --===========================================================================
  architecture ideal of IBIS_VCCS_DELAY is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Iout == Scale * Vin'delayed(TD);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCCS_DELAY is

    generic (TD    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCCS_DELAY;
  --===========================================================================
  architecture ideal of IBIS_CCCS_DELAY is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Iout == Scale * Iin'delayed(TD);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_VCCS_MIN is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCCS_MIN;
  --===========================================================================
  architecture ideal of IBIS_VCCS_MIN is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Iout == Scale * realmin(Vin1, Vin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_CCCS_MIN is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCCS_MIN;
  --===========================================================================
  architecture ideal of IBIS_CCCS_MIN is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Iout == Scale * realmin(Iin1, Iin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_VCCS_MAX is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCCS_MAX;
  --===========================================================================
  architecture ideal of IBIS_VCCS_MAX is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Iout == Scale * realmax(Vin1, Vin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_CCCS_MAX is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCCS_MAX;
  --===========================================================================
  architecture ideal of IBIS_CCCS_MAX is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Iout == Scale * realmax(Iin1, Iin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCCS_ABS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCCS_ABS;
  --===========================================================================
  architecture ideal of IBIS_VCCS_ABS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Iout == Scale * abs(Vin);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCCS_ABS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCCS_ABS;
  --===========================================================================
  architecture ideal of IBIS_CCCS_ABS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Iout == Scale * abs(Iin);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCCS_SUM is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCCS_SUM;
  --===========================================================================
  architecture ideal of IBIS_VCCS_SUM is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Iout == Scale * (Vin1 + Vin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCCS_SUM is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCCS_SUM;
  --===========================================================================
  architecture ideal of IBIS_CCCS_SUM is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Iout == Scale * (Iin1 + Iin2);

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCCS_MULT is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCCS_MULT;
  --===========================================================================
  architecture ideal of IBIS_VCCS_MULT is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Iout == Scale * Vin1 * Vin2;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCCS_MULT is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCCS_MULT;
  --===========================================================================
  architecture ideal of IBIS_CCCS_MULT is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Iout == Scale * Iin1 * Iin2;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCCS_DIV is

    generic (Scale     : real := 1.0;
             ZeroLimit : real := 1.0e-3);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCCS_DIV;
  --===========================================================================
  architecture ideal of IBIS_VCCS_DIV is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    if (Vin2 < -ZeroLimit) or (Vin2 > ZeroLimit) use
      Iout == Scale * Vin1 / Vin2;
    else
      Iout == Scale * Vin1 * Vin2 / ZeroLimit**2;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCCS_DIV is

    generic (Scale     : real := 1.0;
             ZeroLimit : real := 1.0e-3);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCCS_DIV;
  --===========================================================================
  architecture ideal of IBIS_CCCS_DIV is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    if (Iin2 < -ZeroLimit) or (Iin2 > ZeroLimit) use
      Iout == Scale * Iin1 / Iin2;
    else
      Iout == Scale * Iin1 * Iin2 / ZeroLimit**2;
    end use;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_VCCS_PWL is

    generic (Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             X_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
             Y_data : real_vector := (-0.10,  0.00,  0.10,  0.10);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCCS_PWL;
  --===========================================================================
  architecture ideal of IBIS_VCCS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;
  --===========================================================================
  constant  X_vec : real_vector := FileRead(DataFile, "X_data", X_data, Max_length);
  constant  Y_vec : real_vector := FileRead(DataFile, "Y_data", Y_data, Max_length);
--=============================================================================
  begin

    Iout == Scale * Lookup(Vin, X_vec, Y_vec, "SE");

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_CCCS_PWL is

    generic (Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             X_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
             Y_data : real_vector := (-0.10,  0.00,  0.10,  0.10);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCCS_PWL;
  --===========================================================================
  architecture ideal of IBIS_CCCS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;
  --===========================================================================
  constant  X_vec : real_vector := FileRead(DataFile, "X_data", X_data, Max_length);
  constant  Y_vec : real_vector := FileRead(DataFile, "Y_data", Y_data, Max_length);
--=============================================================================
  begin

    Vin  == 0.0;
    Iout == Scale * Lookup(Iin, X_vec, Y_vec, "SE");

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_TCCS_PWL is

    generic (Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             X_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             Y_data : real_vector := (0.00,     0.10,     0.90,     1.00);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n : electrical);

  end entity IBIS_TCCS_PWL;
  --===========================================================================
  architecture ideal of IBIS_TCCS_PWL is

  quantity Vout across Iout through p to n;
  --===========================================================================
  constant  X_vec : real_vector := FileRead(DataFile, "X_data", X_data, Max_length);
  constant  Y_vec : real_vector := FileRead(DataFile, "Y_data", Y_data, Max_length);
--=============================================================================
  begin

    Iout == Scale * Lookup(now, X_vec, Y_vec, "HE");

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_VECCS_PWL is

    generic (Vth_R : real := 0.0;
             Vth_F : real := 0.0;
             Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             XR_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YR_data : real_vector := (0.00,     0.10,     0.90,     1.00);
             XF_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YF_data : real_vector := (1.00,     0.90,     0.10,     0.00);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VECCS_PWL;
  --===========================================================================
  architecture ideal of IBIS_VECCS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;
  signal   T_event    : real    := 0.0;
  signal   EventState : integer := 0;
  --===========================================================================
  constant  XR_vec : real_vector := FileRead(DataFile, "XR_data", XR_data, Max_length);
  constant  YR_vec : real_vector := FileRead(DataFile, "YR_data", YR_data, Max_length);
  constant  XF_vec : real_vector := FileRead(DataFile, "XF_data", XF_data, Max_length);
  constant  YF_vec : real_vector := FileRead(DataFile, "YF_data", YF_data, Max_length);
--=============================================================================
  begin

    ---------------------------------------------------------------------------
    Threshold_crossing : process is
    begin
      if domain = quiescent_domain then
        wait until domain /= quiescent_domain;
        if Vin > Vth_R then            -- high state
          EventState <= 1;
        elsif Vin < Vth_F then         -- low state
          EventState <= -1;
        end if;
      else
        wait on Vin'above(Vth_R), Vin'above(Vth_F);
        if Vin > Vth_R then            -- rising edge
          EventState <= 2;
          T_event    <= now;
        elsif Vin < Vth_F then         -- falling edge
          EventState <= -2;
          T_event    <= now;
        end if;
      end if;
    end process Threshold_crossing;
    ---------------------------------------------------------------------------

    case EventState use
      when -2 =>        -- Executes after falling edge events
        Iout == Scale * Lookup(now-T_event, XF_vec, YF_vec, "HE");
      when -1 =>        -- Executes between t=0 and first threshold crossing
        Iout == Scale * Lookup(0.0, XR_vec, YR_vec, "HE");    -- when input is LOW
      when 1 =>         -- Executes between t=0 and first threshold crossing
        Iout == Scale * Lookup(0.0, XF_vec, YF_vec, "HE");    -- when input is HIGH
      when 2 =>         -- Executes after rising edge events
        Iout == Scale * Lookup(now-T_event, XR_vec, YR_vec, "HE");
      when others =>    -- Executes during DC operating point analysis
        if Vin > Vth_R use
          Iout == Scale * Lookup(0.0, XF_vec, YF_vec, "HE");
        elsif Vin < Vth_F use
          Iout == Scale * Lookup(0.0, XR_vec, YR_vec, "HE");
        else
          Iout == Scale * (Lookup(0.0, XR_vec, YR_vec, "HE") + Lookup(0.0, XF_vec, YF_vec, "HE")) / 2.0;
        end use;
    end case;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_CECCS_PWL is

    generic (Ith_R : real := 0.0;
             Ith_F : real := 0.0;
             Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             XR_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YR_data : real_vector := (0.00,     0.10,     0.90,     1.00);
             XF_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YF_data : real_vector := (1.00,     0.90,     0.10,     0.00);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CECCS_PWL;
  --===========================================================================
  architecture ideal of IBIS_CECCS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;
  signal   T_event    : real    := 0.0;
  signal   EventState : integer := 0;
  --===========================================================================
  constant  XR_vec : real_vector := FileRead(DataFile, "XR_data", XR_data, Max_length);
  constant  YR_vec : real_vector := FileRead(DataFile, "YR_data", YR_data, Max_length);
  constant  XF_vec : real_vector := FileRead(DataFile, "XF_data", XF_data, Max_length);
  constant  YF_vec : real_vector := FileRead(DataFile, "YF_data", YF_data, Max_length);
--=============================================================================
  begin

    ---------------------------------------------------------------------------
    Threshold_crossing : process is
    begin
      if domain = quiescent_domain then
        wait until domain /= quiescent_domain;
        if Iin > Ith_R then            -- high state
          EventState <= 1;
        elsif Iin < Ith_F then         -- low state
          EventState <= -1;
        end if;
      else
        wait on Iin'above(Ith_R), Iin'above(Ith_F);
        if Iin > Ith_R then            -- rising edge
          EventState <= 2;
          T_event    <= now;
        elsif Iin < Ith_F then         -- falling edge
          EventState <= -2;
          T_event    <= now;
        end if;
      end if;
    end process Threshold_crossing;
    ---------------------------------------------------------------------------

    Vin == 0.0;
    case EventState use
      when -2 =>        -- Executes after falling edge events
        Iout == Scale * Lookup(now-T_event, XF_vec, YF_vec, "HE");
      when -1 =>        -- Executes between t=0 and first threshold crossing
        Iout == Scale * Lookup(0.0, XR_vec, YR_vec, "HE");    -- when input is LOW
      when 1 =>         -- Executes between t=0 and first threshold crossing
        Iout == Scale * Lookup(0.0, XF_vec, YF_vec, "HE");    -- when input is HIGH
      when 2 =>         -- Executes after rising edge events
        Iout == Scale * Lookup(now-T_event, XR_vec, YR_vec, "HE");
      when others =>    -- Executes during DC operating point analysis
        if Iin > Ith_R use
          Iout == Scale * Lookup(0.0, XF_vec, YF_vec, "HE");
        elsif Iin < Ith_F use
          Iout == Scale * Lookup(0.0, XR_vec, YR_vec, "HE");
        else
          Iout == Scale * (Lookup(0.0, XR_vec, YR_vec, "HE") + Lookup(0.0, XF_vec, YF_vec, "HE")) / 2.0;
        end use;
    end case;

  end architecture ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_T is

    generic (Z0 : real := 50.0;
             TD : real := 1.0e-9);

    port (terminal n1, ref1, n2, ref2 : electrical);

  end entity IBIS_T;
  --===========================================================================
  architecture Ideal of IBIS_T is

  quantity Vend1 across Iend1 through n1 to ref1;
  quantity Vend2 across Iend2 through n2 to ref2;

  quantity i12 : current;    -- Current wave traveling from end1 to end2
  quantity i21 : current;    -- Current wave traveling from end2 to end1

  begin

    Iend1 == i12 - i21'delayed(TD);
    Iend2 == i21 - i12'delayed(TD);
    Vend1 == Z0 * (2.0 * i12 - Iend1);
    Vend2 == Z0 * (2.0 * i21 - Iend2);

  end architecture Ideal;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_INPUT is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.50;       -- splitting coefficients
           kC_comp_gc : real := 0.50;
           --------------------------------------------------------------------
           -- Receiver thresholds
           --------------------------------------------------------------------
           Vinh : real := 2.00;
           Vinl : real := 0.80;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           MaxIV_length : integer := 100);    -- Maximum possible length of IV tables
     --------------------------------------------------------------------------
     port (terminal PC_ref : electrical;
           terminal GC_ref : electrical;

           terminal Pad    : electrical;
           terminal Rcv_D  : electrical);

end entity IBIS_INPUT;
--=============================================================================
architecture SIMPLE_RECEIVER of IBIS_INPUT is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;
  quantity  Vrcv  across  Ircv  through  Rcv_D   to  ELECTRICAL_REF;

  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vinh_val    : real := FileRead(DataFile, "Vinh", Vinh);
  constant  Vinl_val    : real := FileRead(DataFile, "Vinl", Vinl);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);
--=============================================================================
begin
  --===========================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc == -1.0*kI_pc*Lookup(Vpc, Vpc_vec, Ipc_vec, "SE") + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Igc ==      kI_gc*Lookup(Vgc, Vgc_vec, Igc_vec, "SE") + kC_comp_gc_val*C_comp_val*Vgc'dot;
  -----------------------------------------------------------------------------
  -- A simple receiver logic
  -----------------------------------------------------------------------------
  if    (Vgc > Vinh_val) use Vrcv == 1.0;
  elsif (Vgc < Vinl_val) use Vrcv == 0.0;
  else                       Vrcv == 0.5;
  end use;
--=============================================================================
end architecture SIMPLE_RECEIVER;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_OUTPUT is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.25;       -- splitting coefficients
           kC_comp_pu : real := 0.25;
           kC_comp_pd : real := 0.25;
           kC_comp_gc : real := 0.25;
           --------------------------------------------------------------------
           -- [Pullup Reference] and [Pulldown Reference] values
           --------------------------------------------------------------------
           Vpc_ref : real := 5.0;
           Vpu_ref : real := 5.0;
           Vpd_ref : real := 0.0;
           Vgc_ref : real := 0.0;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Ipu_data : real_vector := ( 0.10,  0.00, -0.10, -0.20);
           Vpu_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Ipd_data : real_vector := (-0.10,  0.00,  0.10,  0.20);
           Vpd_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Vectors of the Vt curve tables
           --------------------------------------------------------------------
           Vr1_data : real_vector := (0.00,  0.00,     2.50,     2.50);
           Tr1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vr2_data : real_vector := (2.50,  2.50,     5.00,     5.00);
           Tr2_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           Vf1_data : real_vector := (5.00,  5.00,     2.50,     2.50);
           Tf1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vf2_data : real_vector := (2.50,  2.50,     0.00,     0.00);
           Tf2_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           --------------------------------------------------------------------
           -- V_fixture and R_fixture values
           --------------------------------------------------------------------
           Vfx_r1 : real := 0.0;
           Vfx_r2 : real := 5.0;
           Vfx_f1 : real := 5.0;
           Vfx_f2 : real := 0.0;

           Rfx_r1 : real := 50.0;
           Rfx_r2 : real := 50.0;
           Rfx_f1 : real := 50.0;
           Rfx_f2 : real := 50.0;
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_pu        : real    := 1.0;     -- PU current scaling
           kI_pd        : real    := 1.0;     -- PD current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling

           kt_rise      : real    := 1.0;     -- Rising waveform time scaling
           kt_fall      : real    := 1.0;     -- Falling waveform time scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           Max_dt       : real    := 1.0e-12; -- Maximum dt between time points
           Vth_R        : real    := 0.8;     -- Input threshold for rising edges
           Vth_F        : real    := 0.2;     -- Input threshold for falling edges
           MaxIV_length : integer := 100;     -- Maximum possible length of IV tables
           MaxVt_length : integer := 1000);   -- Maximum possible length of Vt tables
     --------------------------------------------------------------------------
     port (terminal PU_ref : electrical;
           terminal PD_ref : electrical;

           terminal Pad    : electrical;
           terminal In_D   : electrical;

           terminal PC_ref : electrical;
           terminal GC_ref : electrical);

end entity IBIS_OUTPUT;
--=============================================================================
architecture IBIS_2EQ2UK of IBIS_OUTPUT is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vpu   across  Ipu   through  PU_ref  to  Pad;
  quantity  Vpd   across  Ipd   through  Pad     to  PD_ref;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;

  quantity  Vin   across                 In_D    to  ELECTRICAL_REF;

  signal    pu_on         : integer := 0;
  signal    pu_off        : integer := 0;
  signal    pd_on         : integer := 0;
  signal    pd_off        : integer := 0;
  signal    EventState    : integer := 0;

  quantity  kpu           : current := 0.0;
  quantity  kpd           : current := 0.0;
  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_pu_val : real := FileRead(DataFile, "kC_comp_pu", kC_comp_pu);
  constant  kC_comp_pd_val : real := FileRead(DataFile, "kC_comp_pd", kC_comp_pd);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vpc_ref_val : real := FileRead(DataFile, "Vpc_ref", Vpc_ref);
  constant  Vpu_ref_val : real := FileRead(DataFile, "Vpu_ref", Vpu_ref);
  constant  Vpd_ref_val : real := FileRead(DataFile, "Vpd_ref", Vpd_ref);
  constant  Vgc_ref_val : real := FileRead(DataFile, "Vgc_ref", Vgc_ref);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Ipu_vec     : real_vector := FileRead(DataFile, "Ipu_data", Ipu_data, MaxIV_length);
  constant  Vpu_vec     : real_vector := FileRead(DataFile, "Vpu_data", Vpu_data, MaxIV_length);
  constant  Ipd_vec     : real_vector := FileRead(DataFile, "Ipd_data", Ipd_data, MaxIV_length);
  constant  Vpd_vec     : real_vector := FileRead(DataFile, "Vpd_data", Vpd_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);

  constant  Vr1_vec     : real_vector := FileRead(DataFile, "Vr1_data", Vr1_data, MaxVt_length);
  constant  Tr1_vec     : real_vector := FileRead(DataFile, "Tr1_data", Tr1_data, MaxVt_length, kt_rise);
  constant  Vr2_vec     : real_vector := FileRead(DataFile, "Vr2_data", Vr2_data, MaxVt_length);
  constant  Tr2_vec     : real_vector := FileRead(DataFile, "Tr2_data", Tr2_data, MaxVt_length, kt_rise);
  constant  Vf1_vec     : real_vector := FileRead(DataFile, "Vf1_data", Vf1_data, MaxVt_length);
  constant  Tf1_vec     : real_vector := FileRead(DataFile, "Tf1_data", Tf1_data, MaxVt_length, kt_fall);
  constant  Vf2_vec     : real_vector := FileRead(DataFile, "Vf2_data", Vf2_data, MaxVt_length);
  constant  Tf2_vec     : real_vector := FileRead(DataFile, "Tf2_data", Tf2_data, MaxVt_length, kt_fall);

  constant  Vfx_r1_val  : real := FileRead(DataFile, "Vfx_r1", Vfx_r1);
  constant  Vfx_r2_val  : real := FileRead(DataFile, "Vfx_r2", Vfx_r2);
  constant  Vfx_f1_val  : real := FileRead(DataFile, "Vfx_f1", Vfx_f1);
  constant  Vfx_f2_val  : real := FileRead(DataFile, "Vfx_f2", Vfx_f2);

  constant  Rfx_r1_val  : real := FileRead(DataFile, "Rfx_r1", Rfx_r1);
  constant  Rfx_r2_val  : real := FileRead(DataFile, "Rfx_r2", Rfx_r2);
  constant  Rfx_f1_val  : real := FileRead(DataFile, "Rfx_f1", Rfx_f1);
  constant  Rfx_f2_val  : real := FileRead(DataFile, "Rfx_f2", Rfx_f2);
--=============================================================================
  constant CommonLength : integer := FindCommonLength(Max_dt, Tr1_vec, Tr2_vec, Tf1_vec, Tf2_vec);

  type k_elements    is (Tcm, Kr1, Kr2, Kf1, Kf2);
  type k_tables_type is array(k_elements) of real_vector(1 to CommonLength);
--=============================================================================
  function Coeff (Common_length : integer;
                  Max_dt : real;
                  I_pc   : real_vector;
                  V_pc   : real_vector;
                  I_pu   : real_vector;
                  V_pu   : real_vector;
                  I_pd   : real_vector;
                  V_pd   : real_vector;
                  I_gc   : real_vector;
                  V_gc   : real_vector;

                  V_r1   : real_vector;
                  T_r1   : real_vector;
                  V_r2   : real_vector;
                  T_r2   : real_vector;
                  V_f1   : real_vector;
                  T_f1   : real_vector;
                  V_f2   : real_vector;
                  T_f2   : real_vector;

                  V_pc_ref : real;
                  V_pu_ref : real;
                  V_pd_ref : real;
                  V_gc_ref : real;

                  V_fx_r1  : real;
                  V_fx_r2  : real;
                  V_fx_f1  : real;
                  V_fx_f2  : real;

                  R_fx_r1  : real;
                  R_fx_r2  : real;
                  R_fx_f1  : real;
                  R_fx_f2  : real;

                  C_comp_total : real) return k_tables_type is
  -----------------------------------------------------------------------------
  -- This function converts each Vt table to a corresponding scaling
  -- coefficient table that is used to scale the IV curves with respect
  -- to time.  The C_comp effects in the waveforms are eliminated from the
  -- scaling coefficients, so that the output waveforms will match the Vt
  -- table regardless of the presence of C_comp under identical loading
  -- conditions, i.e. C_comp in the output stage will not derate the
  -- original Vt curves.
  -----------------------------------------------------------------------------
    variable K_tables : k_tables_type;

    variable Vr1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vr2cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf2cm : real_vector(1 to 3) := (others => 0.0);

    variable dVwfm_r1 : real := 0.0;
    variable dVwfm_r2 : real := 0.0;
    variable dVwfm_f1 : real := 0.0;
    variable dVwfm_f2 : real := 0.0;

    variable Iout1 : real := 0.0;
    variable Iout2 : real := 0.0;
    variable Ipc1  : real := 0.0;
    variable Ipc2  : real := 0.0;
    variable Ipu1  : real := 0.0;
    variable Ipu2  : real := 0.0;
    variable Ipd1  : real := 0.0;
    variable Ipd2  : real := 0.0;
    variable Igc1  : real := 0.0;
    variable Igc2  : real := 0.0;

    variable Term1 : real := 0.0;
    variable Term2 : real := 0.0;
    variable Den   : real := 1.0;
  -----------------------------------------------------------------------------
  begin
    K_tables(Tcm) := CommonTime(Common_length, Max_dt, T_r1, T_r2, T_f1, T_f2);
    --=========================================================================
    -- Calculate the scaling coefficients for each time point along Tcm
    ---------------------------------------------------------------------------
    -- Store the first voltage point in the common waveform variables
    ---------------------------------------------------------------------------
    Vr1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r1, V_r1, "HE");
    Vr2cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r2, V_r2, "HE");
    Vf1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f1, V_f1, "HE");
    Vf2cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f2, V_f2, "HE");

    for i in K_tables(Tcm)'range loop
    --report "Tcm: " & real'image(K_tables(Tcm)(i));  -- For debugging
    ---------------------------------------------------------------------------
    -- Store the next point (for the derivative calculations)
    ---------------------------------------------------------------------------
      if (i < K_tables(Tcm)'right) then
        Vr1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r1, V_r1, "HE");
        Vr2cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r2, V_r2, "HE");
        Vf1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f1, V_f1, "HE");
        Vf2cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f2, V_f2, "HE");
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: " & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " R1: " & real'image(Vr1cm(2)) &
      --       " R2: " & real'image(Vr2cm(2)) &
      --       " F1: " & real'image(Vf1cm(2)) &
      --       " F2: " & real'image(Vf2cm(2));
      -------------------------------------------------------------------------
      -- Calculate the derivative of each waveform for C_comp compensation
      -------------------------------------------------------------------------
      if ((i <= K_tables(Tcm)'left) or (i >= K_tables(Tcm)'right)) then
        dVwfm_r1 := 0.0;          -- Force the end point derivatives to zero
        dVwfm_r2 := 0.0;
        dVwfm_f1 := 0.0;
        dVwfm_f2 := 0.0;
      else
        dVwfm_r1 := ((Vr1cm(2) - Vr1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr1cm(3) - Vr1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_r2 := ((Vr2cm(2) - Vr2cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr2cm(3) - Vr2cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f1 := ((Vf1cm(2) - Vf1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf1cm(3) - Vf1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f2 := ((Vf2cm(2) - Vf2cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf2cm(3) - Vf2cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " dR1: " & real'image(dVwfm_r1) &
      --       " dR2: " & real'image(dVwfm_r2) &
      --       " dF1: " & real'image(dVwfm_f1) &
      --       " dF2: " & real'image(dVwfm_f2);
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kr1 Kr2 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vr1cm(2) - V_fx_r1) / R_fx_r1) + C_comp_total * dVwfm_r1;
      Iout2 := ((Vr2cm(2) - V_fx_r2) / R_fx_r2) + C_comp_total * dVwfm_r2;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vr1cm(2), V_pc, I_pc, "HE");
      Ipc2  := -1.0 * Lookup(V_pc_ref - Vr2cm(2), V_pc, I_pc, "HE");
      Ipu1  := -1.0 * Lookup(V_pu_ref - Vr1cm(2), V_pu, I_pu, "HE");
      Ipu2  := -1.0 * Lookup(V_pu_ref - Vr2cm(2), V_pu, I_pu, "HE");
      Ipd1  :=        Lookup(Vr1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Ipd2  :=        Lookup(Vr2cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vr1cm(2) - V_gc_ref, V_gc, I_gc, "HE");
      Igc2  :=        Lookup(Vr2cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      Term1 := Iout1 + Igc1 - Ipc1;
      Term2 := Ipc2  - Igc2 - Iout2;
      Den   := (Ipu1 * Ipd2) - (Ipd1 * Ipu2);

      -- Since these are rising waveforms
      -- Kr1 == Kpu_on and Kr2 == Kpd_off
      K_tables(Kr1)(i) := (Ipd2*Term1 + Ipd1*Term2) / Den;
      K_tables(Kr2)(i) := (Ipu2*Term1 + Ipu1*Term2) / Den;
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kf1 Kf2 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vf1cm(2) - V_fx_f1) / R_fx_f1) + C_comp_total * dVwfm_f1;
      Iout2 := ((Vf2cm(2) - V_fx_f2) / R_fx_f2) + C_comp_total * dVwfm_f2;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vf1cm(2), V_pc, I_pc, "HE");
      Ipc2  := -1.0 * Lookup(V_pc_ref - Vf2cm(2), V_pc, I_pc, "HE");
      Ipu1  := -1.0 * Lookup(V_pu_ref - Vf1cm(2), V_pu, I_pu, "HE");
      Ipu2  := -1.0 * Lookup(V_pu_ref - Vf2cm(2), V_pu, I_pu, "HE");
      Ipd1  :=        Lookup(Vf1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Ipd2  :=        Lookup(Vf2cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vf1cm(2) - V_gc_ref, V_gc, I_gc, "HE");
      Igc2  :=        Lookup(Vf2cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      Term1 := Iout1 + Igc1 - Ipc1;
      Term2 := Ipc2  - Igc2 - Iout2;
      Den   := (Ipu1 * Ipd2) - (Ipd1 * Ipu2);

      -- Since these are falling waveforms
      -- Kf1 == Kpd_on and Kf2 == Kpu_off
      K_tables(Kf1)(i) := (Ipu2*Term1 + Ipu1*Term2) / Den;
      K_tables(Kf2)(i) := (Ipd2*Term1 + Ipd1*Term2) / Den;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " Kr1: " & real'image(K_tables(Kr1)(i)) &
      --       " Kr2: " & real'image(K_tables(Kr2)(i)) &
      --       " Kf1: " & real'image(K_tables(Kf1)(i)) &
      --       " Kf2: " & real'image(K_tables(Kf2)(i));
      -------------------------------------------------------------------------
      -- Shift points by one index for next iteration (for C_comp compensation)
      -------------------------------------------------------------------------
      Vr1cm(1) := Vr1cm(2);
      Vr2cm(1) := Vr2cm(2);
      Vf1cm(1) := Vf1cm(2);
      Vf2cm(1) := Vf2cm(2);

      Vr1cm(2) := Vr1cm(3);
      Vr2cm(2) := Vr2cm(3);
      Vf1cm(2) := Vf1cm(3);
      Vf2cm(2) := Vf2cm(3);

    end loop; -- for loop
    ---------------------------------------------------------------------------
    return K_tables;
  end function Coeff;
--=============================================================================
  constant Ktables : k_tables_type := Coeff(CommonLength, Max_dt,

                                            Ipc_vec, Vpc_vec,
                                            Ipu_vec, Vpu_vec,
                                            Ipd_vec, Vpd_vec,
                                            Igc_vec, Vgc_vec,

                                            Vr1_vec, Tr1_vec,
                                            Vr2_vec, Tr2_vec,
                                            Vf1_vec, Tf1_vec,
                                            Vf2_vec, Tf2_vec,

                                            Vpc_ref_val,
                                            Vpu_ref_val,
                                            Vpd_ref_val,
                                            Vgc_ref_val,

                                            Vfx_r1_val,
                                            Vfx_r2_val,
                                            Vfx_f1_val,
                                            Vfx_f2_val,

                                            Rfx_r1_val,
                                            Rfx_r2_val,
                                            Rfx_f1_val,
                                            Rfx_f2_val,

                                            C_comp_val);
--=============================================================================
begin

  Catch: process is
  -----------------------------------------------------------------------------
  begin
  -----------------------------------------------------------------------------
    if domain = quiescent_domain then

      wait until domain /= quiescent_domain;
      EventState <= 1;

    else

      wait on Vin'above(Vth_R), Vin'above(Vth_F);

      if ((Vin > Vth_R) or (Vin < Vth_F)) then

        if   EventState < 3 then  EventState <= EventState+1;
        else                      EventState <= EventState-1;
        end if;

      end if;
    end if;
  -----------------------------------------------------------------------------
  end process Catch;
  --===========================================================================
  FindState: process (EventState) is
  -----------------------------------------------------------------------------
  begin
  -----------------------------------------------------------------------------
    if (Vin > Vth_R) then                             -- high state
      pu_on  <= 1;
      pd_off <= 1;
      pd_on  <= 0;
      pu_off <= 0;
    elsif (Vin < Vth_F) then                          -- low state
      pu_on  <= 0;
      pd_off <= 0;
      pd_on  <= 1;
      pu_off <= 1;
    else                                              -- unknown state
      pu_on  <= 1;
      pd_off <= 0;
      pd_on  <= 1;
      pu_off <= 0;
    end if;
  -----------------------------------------------------------------------------
  end process FindState;
--=============================================================================
  break on pu_on;
  break on pu_off;
  break on pd_on;
  break on pd_off;
--=============================================================================
  -- This section contains the simultaneous analog equations to find the
  -- appropriate scaling coefficients according to the state the buffer.
  -----------------------------------------------------------------------------
--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
-- This entire section could be made more efficient with nested IF statements
-- if the simulator wouldn't have a problem with simultaneous statements in
-- nested case statements.
--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

  -- Executes during DC operating point calculation
  if    (EventState = 0) and (Vin > Vth_R) use                  -- high state
    kpu == Ktables(Kr1)(CommonLength);
  elsif (EventState = 0) and (Vin < Vth_F) use                  -- low state
    kpu == Ktables(Kf2)(CommonLength);

  -- Executes after DC operating point calculation before first event
  elsif (EventState = 1) and (pu_off = 1) use
    kpu == Ktables(Kf2)(CommonLength);      -- Last point from pu_off table
  elsif (EventState = 1) and (pu_off /= 1)use
    kpu == Ktables(Kr1)(CommonLength);      -- Last point from pu_on table

  -- Executes after first event (normal operation)
  elsif (EventState > 1) and (pu_on  = 1) use    -- Pullup turning on
    kpu == Lookup(pu_on'last_event,  Ktables(Tcm), Ktables(Kr1), "HE");
  elsif (EventState > 1) and (pu_off = 1) use    -- Pullup turning off
    kpu == Lookup(pu_off'last_event, Ktables(Tcm), Ktables(Kf2), "HE");

  -- Just in case
  else
    kpu == Ktables(Kr1)(Ktables(Kr1)'left); -- 1st point of pullup on table
  end use;


  -- Executes during DC operating point calculation
  if    (EventState = 0) and (Vin > Vth_R) use                  -- high state
    kpd == Ktables(Kr2)(CommonLength);
  elsif (EventState = 0) and (Vin < Vth_F) use                  -- low state
    kpd == Ktables(Kf1)(CommonLength);

  -- Executes after DC operating point calculation before first event
  elsif (EventState = 1) and (pd_off = 1) use
    kpd == Ktables(Kr2)(CommonLength);      -- Last point from pd_off table
  elsif (EventState = 1) and (pd_off /= 1)use
    kpd == Ktables(Kf1)(CommonLength);      -- Last point from pd_on table

  -- Executes after first event (normal operation)
  elsif (EventState > 1) and (pd_on  = 1) use    -- Pulldown turning on
    kpd == Lookup(pd_on'last_event,  Ktables(Tcm), Ktables(Kf1), "HE");
  elsif (EventState > 1) and (pd_off = 1) use    -- Pulldown turning off
    kpd == Lookup(pd_off'last_event, Ktables(Tcm), Ktables(Kr2), "HE");

  -- Just in case
  else
    kpd == Ktables(Kf1)(Ktables(Kf1)'left); -- 1st point of pulldown on table
  end use;
--=============================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc == -1.0*kI_pc    *Lookup(Vpc, Vpc_vec, Ipc_vec, "SE");-- + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Ipu == -1.0*kI_pu*kpu*Lookup(Vpu, Vpu_vec, Ipu_vec, "SE");-- + kC_comp_pu_val*C_comp_val*Vpu'dot;
  Ipd ==      kI_pd*kpd*Lookup(Vpd, Vpd_vec, Ipd_vec, "SE");-- + kC_comp_pd_val*C_comp_val*Vpd'dot;
  Igc ==      kI_gc    *Lookup(Vgc, Vgc_vec, Igc_vec, "SE");-- + kC_comp_gc_val*C_comp_val*Vgc'dot;
--=============================================================================
end architecture IBIS_2EQ2UK;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_IO is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.25;       -- splitting coefficients
           kC_comp_pu : real := 0.25;
           kC_comp_pd : real := 0.25;
           kC_comp_gc : real := 0.25;
           --------------------------------------------------------------------
           -- Receiver thresholds
           --------------------------------------------------------------------
           Vinh : real := 2.00;
           Vinl : real := 0.80;
           --------------------------------------------------------------------
           -- [Pullup Reference] and [Pulldown Reference] values
           --------------------------------------------------------------------
           Vpc_ref : real := 5.0;
           Vpu_ref : real := 5.0;
           Vpd_ref : real := 0.0;
           Vgc_ref : real := 0.0;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Ipu_data : real_vector := ( 0.10,  0.00, -0.10, -0.20);
           Vpu_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Ipd_data : real_vector := (-0.10,  0.00,  0.10,  0.20);
           Vpd_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Vectors of the Vt curve tables
           --------------------------------------------------------------------
           Vr1_data : real_vector := (0.00,  0.00,     2.50,     2.50);
           Tr1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vr2_data : real_vector := (2.50,  2.50,     5.00,     5.00);
           Tr2_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           Vf1_data : real_vector := (5.00,  5.00,     2.50,     2.50);
           Tf1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vf2_data : real_vector := (2.50,  2.50,     0.00,     0.00);
           Tf2_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           --------------------------------------------------------------------
           -- V_fixture and R_fixture values
           --------------------------------------------------------------------
           Vfx_r1 : real := 0.0;
           Vfx_r2 : real := 5.0;
           Vfx_f1 : real := 5.0;
           Vfx_f2 : real := 0.0;

           Rfx_r1 : real := 50.0;
           Rfx_r2 : real := 50.0;
           Rfx_f1 : real := 50.0;
           Rfx_f2 : real := 50.0;
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_pu        : real    := 1.0;     -- PU current scaling
           kI_pd        : real    := 1.0;     -- PD current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling

           kt_rise      : real    := 1.0;     -- Rising waveform time scaling
           kt_fall      : real    := 1.0;     -- Falling waveform time scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           Max_dt       : real    := 1.0e-12; -- Maximum dt between time points
           Vth_R        : real    := 0.8;     -- Input threshold for rising edges
           Vth_F        : real    := 0.2;     -- Input threshold for falling edges
           MaxIV_length : integer := 100;     -- Maximum possible length of IV tables
           MaxVt_length : integer := 1000);   -- Maximum possible length of Vt tables
     --------------------------------------------------------------------------
     port (terminal PU_ref : electrical;
           terminal PD_ref : electrical;

           terminal Pad    : electrical;
           terminal In_D   : electrical;
           terminal En_D   : electrical;
--           terminal Rcv_D  : electrical;

           terminal PC_ref : electrical;
           terminal GC_ref : electrical);

end entity IBIS_IO;
--=============================================================================
architecture IBIS_2EQ2UK of IBIS_IO is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vpu   across  Ipu   through  PU_ref  to  Pad;
  quantity  Vpd   across  Ipd   through  Pad     to  PD_ref;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;

  quantity  Vin   across                 In_D    to  ELECTRICAL_REF;
  quantity  Ven   across                 En_D    to  ELECTRICAL_REF;
--  quantity  Vrcv  across  Ircv  through  Rcv_D   to  ELECTRICAL_REF;

  signal    pu_on         : integer := 0;
  signal    pu_off        : integer := 0;
  signal    pd_on         : integer := 0;
  signal    pd_off        : integer := 0;
  signal    EventState    : integer := 0;

  quantity  kpu           : current := 0.0;
  quantity  kpd           : current := 0.0;
  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_pu_val : real := FileRead(DataFile, "kC_comp_pu", kC_comp_pu);
  constant  kC_comp_pd_val : real := FileRead(DataFile, "kC_comp_pd", kC_comp_pd);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vinh_val    : real := FileRead(DataFile, "Vinh", Vinh);
  constant  Vinl_val    : real := FileRead(DataFile, "Vinl", Vinl);

  constant  Vpc_ref_val : real := FileRead(DataFile, "Vpc_ref", Vpc_ref);
  constant  Vpu_ref_val : real := FileRead(DataFile, "Vpu_ref", Vpu_ref);
  constant  Vpd_ref_val : real := FileRead(DataFile, "Vpd_ref", Vpd_ref);
  constant  Vgc_ref_val : real := FileRead(DataFile, "Vgc_ref", Vgc_ref);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Ipu_vec     : real_vector := FileRead(DataFile, "Ipu_data", Ipu_data, MaxIV_length);
  constant  Vpu_vec     : real_vector := FileRead(DataFile, "Vpu_data", Vpu_data, MaxIV_length);
  constant  Ipd_vec     : real_vector := FileRead(DataFile, "Ipd_data", Ipd_data, MaxIV_length);
  constant  Vpd_vec     : real_vector := FileRead(DataFile, "Vpd_data", Vpd_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);

  constant  Vr1_vec     : real_vector := FileRead(DataFile, "Vr1_data", Vr1_data, MaxVt_length);
  constant  Tr1_vec     : real_vector := FileRead(DataFile, "Tr1_data", Tr1_data, MaxVt_length, kt_rise);
  constant  Vr2_vec     : real_vector := FileRead(DataFile, "Vr2_data", Vr2_data, MaxVt_length);
  constant  Tr2_vec     : real_vector := FileRead(DataFile, "Tr2_data", Tr2_data, MaxVt_length, kt_rise);
  constant  Vf1_vec     : real_vector := FileRead(DataFile, "Vf1_data", Vf1_data, MaxVt_length);
  constant  Tf1_vec     : real_vector := FileRead(DataFile, "Tf1_data", Tf1_data, MaxVt_length, kt_fall);
  constant  Vf2_vec     : real_vector := FileRead(DataFile, "Vf2_data", Vf2_data, MaxVt_length);
  constant  Tf2_vec     : real_vector := FileRead(DataFile, "Tf2_data", Tf2_data, MaxVt_length, kt_fall);

  constant  Vfx_r1_val  : real := FileRead(DataFile, "Vfx_r1", Vfx_r1);
  constant  Vfx_r2_val  : real := FileRead(DataFile, "Vfx_r2", Vfx_r2);
  constant  Vfx_f1_val  : real := FileRead(DataFile, "Vfx_f1", Vfx_f1);
  constant  Vfx_f2_val  : real := FileRead(DataFile, "Vfx_f2", Vfx_f2);

  constant  Rfx_r1_val  : real := FileRead(DataFile, "Rfx_r1", Rfx_r1);
  constant  Rfx_r2_val  : real := FileRead(DataFile, "Rfx_r2", Rfx_r2);
  constant  Rfx_f1_val  : real := FileRead(DataFile, "Rfx_f1", Rfx_f1);
  constant  Rfx_f2_val  : real := FileRead(DataFile, "Rfx_f2", Rfx_f2);
--=============================================================================
  constant CommonLength : integer := FindCommonLength(Max_dt, Tr1_vec, Tr2_vec, Tf1_vec, Tf2_vec);

  type k_elements    is (Tcm, Kr1, Kr2, Kf1, Kf2);
  type k_tables_type is array(k_elements) of real_vector(1 to CommonLength);
--=============================================================================
  function Coeff (Common_length : integer;
                  Max_dt : real;
                  I_pc   : real_vector;
                  V_pc   : real_vector;
                  I_pu   : real_vector;
                  V_pu   : real_vector;
                  I_pd   : real_vector;
                  V_pd   : real_vector;
                  I_gc   : real_vector;
                  V_gc   : real_vector;
  
                  V_r1   : real_vector;
                  T_r1   : real_vector;
                  V_r2   : real_vector;
                  T_r2   : real_vector;
                  V_f1   : real_vector;
                  T_f1   : real_vector;
                  V_f2   : real_vector;
                  T_f2   : real_vector;
  
                  V_pc_ref : real;
                  V_pu_ref : real;
                  V_pd_ref : real;
                  V_gc_ref : real;
  
                  V_fx_r1  : real;
                  V_fx_r2  : real;
                  V_fx_f1  : real;
                  V_fx_f2  : real;
  
                  R_fx_r1  : real;
                  R_fx_r2  : real;
                  R_fx_f1  : real;
                  R_fx_f2  : real;
  
                  C_comp_total : real) return k_tables_type is
  -----------------------------------------------------------------------------
  -- This function converts each Vt table to a corresponding scaling
  -- coefficient table that is used to scale the IV curves with respect
  -- to time.  The C_comp effects in the waveforms are eliminated from the
  -- scaling coefficients, so that the output waveforms will match the Vt
  -- table regardless of the presence of C_comp under identical loading
  -- conditions, i.e. C_comp in the output stage will not derate the
  -- original Vt curves.
  -----------------------------------------------------------------------------
    variable K_tables : k_tables_type;
  
    variable Vr1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vr2cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf2cm : real_vector(1 to 3) := (others => 0.0);
  
    variable dVwfm_r1 : real := 0.0;
    variable dVwfm_r2 : real := 0.0;
    variable dVwfm_f1 : real := 0.0;
    variable dVwfm_f2 : real := 0.0;
  
    variable Iout1 : real := 0.0;
    variable Iout2 : real := 0.0;
    variable Ipc1  : real := 0.0;
    variable Ipc2  : real := 0.0;
    variable Ipu1  : real := 0.0;
    variable Ipu2  : real := 0.0;
    variable Ipd1  : real := 0.0;
    variable Ipd2  : real := 0.0;
    variable Igc1  : real := 0.0;
    variable Igc2  : real := 0.0;
  
    variable Term1 : real := 0.0;
    variable Term2 : real := 0.0;
    variable Den   : real := 1.0;
  -----------------------------------------------------------------------------
  begin
    K_tables(Tcm) := CommonTime(Common_length, Max_dt, T_r1, T_r2, T_f1, T_f2);
    --=========================================================================
    -- Calculate the scaling coefficients for each time point along Tcm
    ---------------------------------------------------------------------------
    -- Store the first voltage point in the common waveform variables
    ---------------------------------------------------------------------------
    Vr1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r1, V_r1, "HE");
    Vr2cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r2, V_r2, "HE");
    Vf1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f1, V_f1, "HE");
    Vf2cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f2, V_f2, "HE");
  
    for i in K_tables(Tcm)'range loop
    --report "Tcm: " & real'image(K_tables(Tcm)(i));  -- For debugging
    ---------------------------------------------------------------------------
    -- Store the next point (for the derivative calculations)
    ---------------------------------------------------------------------------
      if (i < K_tables(Tcm)'right) then
        Vr1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r1, V_r1, "HE");
        Vr2cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r2, V_r2, "HE");
        Vf1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f1, V_f1, "HE");
        Vf2cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f2, V_f2, "HE");
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: " & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " R1: " & real'image(Vr1cm(2)) &
      --       " R2: " & real'image(Vr2cm(2)) &
      --       " F1: " & real'image(Vf1cm(2)) &
      --       " F2: " & real'image(Vf2cm(2));
      -------------------------------------------------------------------------
      -- Calculate the derivative of each waveform for C_comp compensation
      -------------------------------------------------------------------------
      if ((i <= K_tables(Tcm)'left) or (i >= K_tables(Tcm)'right)) then
        dVwfm_r1 := 0.0;          -- Force the end point derivatives to zero
        dVwfm_r2 := 0.0;
        dVwfm_f1 := 0.0;
        dVwfm_f2 := 0.0;
      else
        dVwfm_r1 := ((Vr1cm(2) - Vr1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr1cm(3) - Vr1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_r2 := ((Vr2cm(2) - Vr2cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr2cm(3) - Vr2cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f1 := ((Vf1cm(2) - Vf1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf1cm(3) - Vf1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f2 := ((Vf2cm(2) - Vf2cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf2cm(3) - Vf2cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " dR1: " & real'image(dVwfm_r1) &
      --       " dR2: " & real'image(dVwfm_r2) &
      --       " dF1: " & real'image(dVwfm_f1) &
      --       " dF2: " & real'image(dVwfm_f2);
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kr1 Kr2 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vr1cm(2) - V_fx_r1) / R_fx_r1) + C_comp_total * dVwfm_r1;
      Iout2 := ((Vr2cm(2) - V_fx_r2) / R_fx_r2) + C_comp_total * dVwfm_r2;
  
      Ipc1  := -1.0 * Lookup(V_pc_ref - Vr1cm(2), V_pc, I_pc, "HE");
      Ipc2  := -1.0 * Lookup(V_pc_ref - Vr2cm(2), V_pc, I_pc, "HE");
      Ipu1  := -1.0 * Lookup(V_pu_ref - Vr1cm(2), V_pu, I_pu, "HE");
      Ipu2  := -1.0 * Lookup(V_pu_ref - Vr2cm(2), V_pu, I_pu, "HE");
      Ipd1  :=        Lookup(Vr1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Ipd2  :=        Lookup(Vr2cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vr1cm(2) - V_gc_ref, V_gc, I_gc, "HE");
      Igc2  :=        Lookup(Vr2cm(2) - V_gc_ref, V_gc, I_gc, "HE");
  
      Term1 := Iout1 + Igc1 - Ipc1;
      Term2 := Ipc2  - Igc2 - Iout2;
      Den   := (Ipu1 * Ipd2) - (Ipd1 * Ipu2);
  
      -- Since these are rising waveforms
      -- Kr1 == Kpu_on and Kr2 == Kpd_off
      K_tables(Kr1)(i) := (Ipd2*Term1 + Ipd1*Term2) / Den;
      K_tables(Kr2)(i) := (Ipu2*Term1 + Ipu1*Term2) / Den;
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kf1 Kf2 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vf1cm(2) - V_fx_f1) / R_fx_f1) + C_comp_total * dVwfm_f1;
      Iout2 := ((Vf2cm(2) - V_fx_f2) / R_fx_f2) + C_comp_total * dVwfm_f2;
  
      Ipc1  := -1.0 * Lookup(V_pc_ref - Vf1cm(2), V_pc, I_pc, "HE");
      Ipc2  := -1.0 * Lookup(V_pc_ref - Vf2cm(2), V_pc, I_pc, "HE");
      Ipu1  := -1.0 * Lookup(V_pu_ref - Vf1cm(2), V_pu, I_pu, "HE");
      Ipu2  := -1.0 * Lookup(V_pu_ref - Vf2cm(2), V_pu, I_pu, "HE");
      Ipd1  :=        Lookup(Vf1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Ipd2  :=        Lookup(Vf2cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vf1cm(2) - V_gc_ref, V_gc, I_gc, "HE");
      Igc2  :=        Lookup(Vf2cm(2) - V_gc_ref, V_gc, I_gc, "HE");
  
      Term1 := Iout1 + Igc1 - Ipc1;
      Term2 := Ipc2  - Igc2 - Iout2;
      Den   := (Ipu1 * Ipd2) - (Ipd1 * Ipu2);
  
      -- Since these are falling waveforms
      -- Kf1 == Kpd_on and Kf2 == Kpu_off
      K_tables(Kf1)(i) := (Ipu2*Term1 + Ipu1*Term2) / Den;
      K_tables(Kf2)(i) := (Ipd2*Term1 + Ipd1*Term2) / Den;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " Kr1: " & real'image(K_tables(Kr1)(i)) &
      --       " Kr2: " & real'image(K_tables(Kr2)(i)) &
      --       " Kf1: " & real'image(K_tables(Kf1)(i)) &
      --       " Kf2: " & real'image(K_tables(Kf2)(i));
      -------------------------------------------------------------------------
      -- Shift points by one index for next iteration (for C_comp compensation)
      -------------------------------------------------------------------------
      Vr1cm(1) := Vr1cm(2);
      Vr2cm(1) := Vr2cm(2);
      Vf1cm(1) := Vf1cm(2);
      Vf2cm(1) := Vf2cm(2);
  
      Vr1cm(2) := Vr1cm(3);
      Vr2cm(2) := Vr2cm(3);
      Vf1cm(2) := Vf1cm(3);
      Vf2cm(2) := Vf2cm(3);
  
    end loop; -- for loop
    ---------------------------------------------------------------------------
    return K_tables;
  end function Coeff;
--=============================================================================
  constant Ktables : k_tables_type := Coeff(CommonLength, Max_dt,

                                            Ipc_vec, Vpc_vec,
                                            Ipu_vec, Vpu_vec,
                                            Ipd_vec, Vpd_vec,
                                            Igc_vec, Vgc_vec,

                                            Vr1_vec, Tr1_vec,
                                            Vr2_vec, Tr2_vec,
                                            Vf1_vec, Tf1_vec,
                                            Vf2_vec, Tf2_vec,

                                            Vpc_ref_val,
                                            Vpu_ref_val,
                                            Vpd_ref_val,
                                            Vgc_ref_val,

                                            Vfx_r1_val,
                                            Vfx_r2_val,
                                            Vfx_f1_val,
                                            Vfx_f2_val,

                                            Rfx_r1_val,
                                            Rfx_r2_val,
                                            Rfx_f1_val,
                                            Rfx_f2_val,

                                            C_comp_val);
--=============================================================================
begin

  Catch: process is
  -----------------------------------------------------------------------------
  begin
  -----------------------------------------------------------------------------
    if domain = quiescent_domain then

      wait until domain /= quiescent_domain;
      EventState <= 1;

    else

      wait on Vin'above(Vth_R), Vin'above(Vth_F),
              Ven'above(Vth_R), Ven'above(Vth_F);

      if ((Vin > Vth_R) or (Vin < Vth_F)) and
         ((Ven > Vth_R) or (Ven < Vth_F)) then

        if   EventState < 3 then  EventState <= EventState+1;
        else                      EventState <= EventState-1;
        end if;

      end if;
    end if;
  -----------------------------------------------------------------------------
  end process Catch;
  --===========================================================================
  FindState: process (EventState) is
  -----------------------------------------------------------------------------
  begin
  -----------------------------------------------------------------------------
    if (Ven > Vth_R) and (Vin > Vth_R) then           -- high state
      pu_on  <= 1;
      pd_off <= 1;
      pd_on  <= 0;
      pu_off <= 0;
    elsif (Ven > Vth_R) and (Vin < Vth_F) then        -- low state
      pu_on  <= 0;
      pd_off <= 0;
      pd_on  <= 1;
      pu_off <= 1;
    elsif (Ven < Vth_F) then                          -- 3-state
      pu_on  <= 0;
      pd_off <= 1;
      pd_on  <= 0;
      pu_off <= 1;
    else                                              -- unknown state
      pu_on  <= 1;
      pd_off <= 0;
      pd_on  <= 1;
      pu_off <= 0;
    end if;
  -----------------------------------------------------------------------------
  end process FindState;
--=============================================================================
  break on pu_on;
  break on pu_off;
  break on pd_on;
  break on pd_off;
--=============================================================================
  -- This section contains the simultaneous analog equations to find the
  -- appropriate scaling coefficients according to the state the buffer.
  -----------------------------------------------------------------------------
--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
-- This entire section could be made more efficient with nested IF statements
-- if the simulator wouldn't have a problem with simultaneous statements in
-- nested case statements.
--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

  -- Executes during DC operating point calculation
  if    (EventState = 0) and (Ven > Vth_R) and (Vin > Vth_R) use   -- high state
    kpu == Ktables(Kr1)(CommonLength);
  elsif (EventState = 0) and (Ven > Vth_R) and (Vin < Vth_F) use   -- low state
    kpu == Ktables(Kf2)(CommonLength);
  elsif (EventState = 0) and (Ven < Vth_F) use                     -- 3-state
    kpu == Ktables(Kf2)(CommonLength);

  -- Executes after DC operating point calculation before first event
  elsif (EventState = 1) and (pu_off = 1) use
    kpu == Ktables(Kf2)(CommonLength);      -- Last point from pu_off table
  elsif (EventState = 1) and (pu_off /= 1)use
    kpu == Ktables(Kr1)(CommonLength);      -- Last point from pu_on table

  -- Executes after first event (normal operation)
  elsif (EventState > 1) and (pu_on  = 1) use    -- Pullup turning on
    kpu == Lookup(pu_on'last_event,  Ktables(Tcm), Ktables(Kr1), "HE");
  elsif (EventState > 1) and (pu_off = 1) use    -- Pullup turning off
    kpu == Lookup(pu_off'last_event, Ktables(Tcm), Ktables(Kf2), "HE");

  -- Just in case
  else
    kpu == Ktables(Kr1)(Ktables(Kr1)'left); -- 1st point of pullup on table
  end use;


  -- Executes during DC operating point calculation
  if    (EventState = 0) and (Ven > Vth_R) and (Vin > Vth_R) use   -- high state
    kpd == Ktables(Kr2)(CommonLength);
  elsif (EventState = 0) and (Ven > Vth_R) and (Vin < Vth_F) use   -- low state
    kpd == Ktables(Kf1)(CommonLength);
  elsif (EventState = 0) and (Ven < Vth_F) use                     -- 3-state
    kpd == Ktables(Kr2)(CommonLength);

  -- Executes after DC operating point calculation before first event
  elsif (EventState = 1) and (pd_off = 1) use
    kpd == Ktables(Kr2)(CommonLength);      -- Last point from pd_off table
  elsif (EventState = 1) and (pd_off /= 1)use
    kpd == Ktables(Kf1)(CommonLength);      -- Last point from pd_on table

  -- Executes after first event (normal operation)
  elsif (EventState > 1) and (pd_on  = 1) use    -- Pulldown turning on
    kpd == Lookup(pd_on'last_event,  Ktables(Tcm), Ktables(Kf1), "HE");
  elsif (EventState > 1) and (pd_off = 1) use    -- Pulldown turning off
    kpd == Lookup(pd_off'last_event, Ktables(Tcm), Ktables(Kr2), "HE");

  -- Just in case
  else
    kpd == Ktables(Kf1)(Ktables(Kf1)'left); -- 1st point of pulldown on table
  end use;
--=============================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc   == -1.0*kI_pc    *Lookup(Vpc, Vpc_vec, Ipc_vec, "SE");-- + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Ipu   == -1.0*kI_pu*kpu*Lookup(Vpu, Vpu_vec, Ipu_vec, "SE");-- + kC_comp_pu_val*C_comp_val*Vpu'dot;
  Ipd   ==      kI_pd*kpd*Lookup(Vpd, Vpd_vec, Ipd_vec, "SE");-- + kC_comp_pd_val*C_comp_val*Vpd'dot;
  Igc   ==      kI_gc    *Lookup(Vgc, Vgc_vec, Igc_vec, "SE");-- + kC_comp_gc_val*C_comp_val*Vgc'dot;
  -----------------------------------------------------------------------------
  -- A simple receiver logic
  -----------------------------------------------------------------------------
--  if    (Vpd > Vinh_val) use Vrcv == 1.0;
--  elsif (Vpd < Vinl_val) use Vrcv == 0.0;
--  else                       Vrcv == 0.5;
--  end use;
--=============================================================================
end architecture IBIS_2EQ2UK;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_3STATE is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.25;       -- splitting coefficients
           kC_comp_pu : real := 0.25;
           kC_comp_pd : real := 0.25;
           kC_comp_gc : real := 0.25;
           --------------------------------------------------------------------
           -- [Pullup Reference] and [Pulldown Reference] values
           --------------------------------------------------------------------
           Vpc_ref : real := 5.0;
           Vpu_ref : real := 5.0;
           Vpd_ref : real := 0.0;
           Vgc_ref : real := 0.0;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Ipu_data : real_vector := ( 0.10,  0.00, -0.10, -0.20);
           Vpu_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Ipd_data : real_vector := (-0.10,  0.00,  0.10,  0.20);
           Vpd_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Vectors of the Vt curve tables
           --------------------------------------------------------------------
           Vr1_data : real_vector := (0.00,  0.00,     2.50,     2.50);
           Tr1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vr2_data : real_vector := (2.50,  2.50,     5.00,     5.00);
           Tr2_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           Vf1_data : real_vector := (5.00,  5.00,     2.50,     2.50);
           Tf1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vf2_data : real_vector := (2.50,  2.50,     0.00,     0.00);
           Tf2_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           --------------------------------------------------------------------
           -- V_fixture and R_fixture values
           --------------------------------------------------------------------
           Vfx_r1 : real := 0.0;
           Vfx_r2 : real := 5.0;
           Vfx_f1 : real := 5.0;
           Vfx_f2 : real := 0.0;

           Rfx_r1 : real := 50.0;
           Rfx_r2 : real := 50.0;
           Rfx_f1 : real := 50.0;
           Rfx_f2 : real := 50.0;
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_pu        : real    := 1.0;     -- PU current scaling
           kI_pd        : real    := 1.0;     -- PD current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling

           kt_rise      : real    := 1.0;     -- Rising waveform time scaling
           kt_fall      : real    := 1.0;     -- Falling waveform time scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           Max_dt       : real    := 1.0e-12; -- Maximum dt between time points
           Vth_R        : real    := 0.8;     -- Input threshold for rising edges
           Vth_F        : real    := 0.2;     -- Input threshold for falling edges
           MaxIV_length : integer := 100;     -- Maximum possible length of IV tables
           MaxVt_length : integer := 1000);   -- Maximum possible length of Vt tables
     --------------------------------------------------------------------------
     port (terminal PU_ref : electrical;
           terminal PD_ref : electrical;

           terminal Pad    : electrical;
           terminal In_D   : electrical;
           terminal En_D   : electrical;

           terminal PC_ref : electrical;
           terminal GC_ref : electrical);

end entity IBIS_3STATE;
--=============================================================================
architecture IBIS_2EQ2UK of IBIS_3STATE is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vpu   across  Ipu   through  PU_ref  to  Pad;
  quantity  Vpd   across  Ipd   through  Pad     to  PD_ref;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;

  quantity  Vin   across                 In_D    to  ELECTRICAL_REF;
  quantity  Ven   across                 En_D    to  ELECTRICAL_REF;

  signal    pu_on         : integer := 0;
  signal    pu_off        : integer := 0;
  signal    pd_on         : integer := 0;
  signal    pd_off        : integer := 0;
  signal    EventState    : integer := 0;

  quantity  kpu           : current := 0.0;
  quantity  kpd           : current := 0.0;
  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_pu_val : real := FileRead(DataFile, "kC_comp_pu", kC_comp_pu);
  constant  kC_comp_pd_val : real := FileRead(DataFile, "kC_comp_pd", kC_comp_pd);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vpc_ref_val : real := FileRead(DataFile, "Vpc_ref", Vpc_ref);
  constant  Vpu_ref_val : real := FileRead(DataFile, "Vpu_ref", Vpu_ref);
  constant  Vpd_ref_val : real := FileRead(DataFile, "Vpd_ref", Vpd_ref);
  constant  Vgc_ref_val : real := FileRead(DataFile, "Vgc_ref", Vgc_ref);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Ipu_vec     : real_vector := FileRead(DataFile, "Ipu_data", Ipu_data, MaxIV_length);
  constant  Vpu_vec     : real_vector := FileRead(DataFile, "Vpu_data", Vpu_data, MaxIV_length);
  constant  Ipd_vec     : real_vector := FileRead(DataFile, "Ipd_data", Ipd_data, MaxIV_length);
  constant  Vpd_vec     : real_vector := FileRead(DataFile, "Vpd_data", Vpd_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);

  constant  Vr1_vec     : real_vector := FileRead(DataFile, "Vr1_data", Vr1_data, MaxVt_length);
  constant  Tr1_vec     : real_vector := FileRead(DataFile, "Tr1_data", Tr1_data, MaxVt_length, kt_rise);
  constant  Vr2_vec     : real_vector := FileRead(DataFile, "Vr2_data", Vr2_data, MaxVt_length);
  constant  Tr2_vec     : real_vector := FileRead(DataFile, "Tr2_data", Tr2_data, MaxVt_length, kt_rise);
  constant  Vf1_vec     : real_vector := FileRead(DataFile, "Vf1_data", Vf1_data, MaxVt_length);
  constant  Tf1_vec     : real_vector := FileRead(DataFile, "Tf1_data", Tf1_data, MaxVt_length, kt_fall);
  constant  Vf2_vec     : real_vector := FileRead(DataFile, "Vf2_data", Vf2_data, MaxVt_length);
  constant  Tf2_vec     : real_vector := FileRead(DataFile, "Tf2_data", Tf2_data, MaxVt_length, kt_fall);

  constant  Vfx_r1_val  : real := FileRead(DataFile, "Vfx_r1", Vfx_r1);
  constant  Vfx_r2_val  : real := FileRead(DataFile, "Vfx_r2", Vfx_r2);
  constant  Vfx_f1_val  : real := FileRead(DataFile, "Vfx_f1", Vfx_f1);
  constant  Vfx_f2_val  : real := FileRead(DataFile, "Vfx_f2", Vfx_f2);

  constant  Rfx_r1_val  : real := FileRead(DataFile, "Rfx_r1", Rfx_r1);
  constant  Rfx_r2_val  : real := FileRead(DataFile, "Rfx_r2", Rfx_r2);
  constant  Rfx_f1_val  : real := FileRead(DataFile, "Rfx_f1", Rfx_f1);
  constant  Rfx_f2_val  : real := FileRead(DataFile, "Rfx_f2", Rfx_f2);
--=============================================================================
  constant CommonLength : integer := FindCommonLength(Max_dt, Tr1_vec, Tr2_vec, Tf1_vec, Tf2_vec);

  type k_elements    is (Tcm, Kr1, Kr2, Kf1, Kf2);
  type k_tables_type is array(k_elements) of real_vector(1 to CommonLength);
--=============================================================================
  function Coeff (Common_length : integer;
                  Max_dt : real;
                  I_pc   : real_vector;
                  V_pc   : real_vector;
                  I_pu   : real_vector;
                  V_pu   : real_vector;
                  I_pd   : real_vector;
                  V_pd   : real_vector;
                  I_gc   : real_vector;
                  V_gc   : real_vector;

                  V_r1   : real_vector;
                  T_r1   : real_vector;
                  V_r2   : real_vector;
                  T_r2   : real_vector;
                  V_f1   : real_vector;
                  T_f1   : real_vector;
                  V_f2   : real_vector;
                  T_f2   : real_vector;

                  V_pc_ref : real;
                  V_pu_ref : real;
                  V_pd_ref : real;
                  V_gc_ref : real;

                  V_fx_r1  : real;
                  V_fx_r2  : real;
                  V_fx_f1  : real;
                  V_fx_f2  : real;

                  R_fx_r1  : real;
                  R_fx_r2  : real;
                  R_fx_f1  : real;
                  R_fx_f2  : real;

                  C_comp_total : real) return k_tables_type is
  -----------------------------------------------------------------------------
  -- This function converts each Vt table to a corresponding scaling
  -- coefficient table that is used to scale the IV curves with respect
  -- to time.  The C_comp effects in the waveforms are eliminated from the
  -- scaling coefficients, so that the output waveforms will match the Vt
  -- table regardless of the presence of C_comp under identical loading
  -- conditions, i.e. C_comp in the output stage will not derate the
  -- original Vt curves.
  -----------------------------------------------------------------------------
    variable K_tables : k_tables_type;

    variable Vr1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vr2cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf2cm : real_vector(1 to 3) := (others => 0.0);

    variable dVwfm_r1 : real := 0.0;
    variable dVwfm_r2 : real := 0.0;
    variable dVwfm_f1 : real := 0.0;
    variable dVwfm_f2 : real := 0.0;

    variable Iout1 : real := 0.0;
    variable Iout2 : real := 0.0;
    variable Ipc1  : real := 0.0;
    variable Ipc2  : real := 0.0;
    variable Ipu1  : real := 0.0;
    variable Ipu2  : real := 0.0;
    variable Ipd1  : real := 0.0;
    variable Ipd2  : real := 0.0;
    variable Igc1  : real := 0.0;
    variable Igc2  : real := 0.0;

    variable Term1 : real := 0.0;
    variable Term2 : real := 0.0;
    variable Den   : real := 1.0;
  -----------------------------------------------------------------------------
  begin
    K_tables(Tcm) := CommonTime(Common_length, Max_dt, T_r1, T_r2, T_f1, T_f2);
    --=========================================================================
    -- Calculate the scaling coefficients for each time point along Tcm
    ---------------------------------------------------------------------------
    -- Store the first voltage point in the common waveform variables
    ---------------------------------------------------------------------------
    Vr1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r1, V_r1, "HE");
    Vr2cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r2, V_r2, "HE");
    Vf1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f1, V_f1, "HE");
    Vf2cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f2, V_f2, "HE");

    for i in K_tables(Tcm)'range loop
    --report "Tcm: " & real'image(K_tables(Tcm)(i));  -- For debugging
    ---------------------------------------------------------------------------
    -- Store the next point (for the derivative calculations)
    ---------------------------------------------------------------------------
      if (i < K_tables(Tcm)'right) then
        Vr1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r1, V_r1, "HE");
        Vr2cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r2, V_r2, "HE");
        Vf1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f1, V_f1, "HE");
        Vf2cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f2, V_f2, "HE");
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: " & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " R1: " & real'image(Vr1cm(2)) &
      --       " R2: " & real'image(Vr2cm(2)) &
      --       " F1: " & real'image(Vf1cm(2)) &
      --       " F2: " & real'image(Vf2cm(2));
      -------------------------------------------------------------------------
      -- Calculate the derivative of each waveform for C_comp compensation
      -------------------------------------------------------------------------
      if ((i <= K_tables(Tcm)'left) or (i >= K_tables(Tcm)'right)) then
        dVwfm_r1 := 0.0;          -- Force the end point derivatives to zero
        dVwfm_r2 := 0.0;
        dVwfm_f1 := 0.0;
        dVwfm_f2 := 0.0;
      else
        dVwfm_r1 := ((Vr1cm(2) - Vr1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr1cm(3) - Vr1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_r2 := ((Vr2cm(2) - Vr2cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr2cm(3) - Vr2cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f1 := ((Vf1cm(2) - Vf1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf1cm(3) - Vf1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f2 := ((Vf2cm(2) - Vf2cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf2cm(3) - Vf2cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " dR1: " & real'image(dVwfm_r1) &
      --       " dR2: " & real'image(dVwfm_r2) &
      --       " dF1: " & real'image(dVwfm_f1) &
      --       " dF2: " & real'image(dVwfm_f2);
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kr1 Kr2 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vr1cm(2) - V_fx_r1) / R_fx_r1) + C_comp_total * dVwfm_r1;
      Iout2 := ((Vr2cm(2) - V_fx_r2) / R_fx_r2) + C_comp_total * dVwfm_r2;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vr1cm(2), V_pc, I_pc, "HE");
      Ipc2  := -1.0 * Lookup(V_pc_ref - Vr2cm(2), V_pc, I_pc, "HE");
      Ipu1  := -1.0 * Lookup(V_pu_ref - Vr1cm(2), V_pu, I_pu, "HE");
      Ipu2  := -1.0 * Lookup(V_pu_ref - Vr2cm(2), V_pu, I_pu, "HE");
      Ipd1  :=        Lookup(Vr1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Ipd2  :=        Lookup(Vr2cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vr1cm(2) - V_gc_ref, V_gc, I_gc, "HE");
      Igc2  :=        Lookup(Vr2cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      Term1 := Iout1 + Igc1 - Ipc1;
      Term2 := Ipc2  - Igc2 - Iout2;
      Den   := (Ipu1 * Ipd2) - (Ipd1 * Ipu2);

      -- Since these are rising waveforms
      -- Kr1 == Kpu_on and Kr2 == Kpd_off
      K_tables(Kr1)(i) := (Ipd2*Term1 + Ipd1*Term2) / Den;
      K_tables(Kr2)(i) := (Ipu2*Term1 + Ipu1*Term2) / Den;
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kf1 Kf2 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vf1cm(2) - V_fx_f1) / R_fx_f1) + C_comp_total * dVwfm_f1;
      Iout2 := ((Vf2cm(2) - V_fx_f2) / R_fx_f2) + C_comp_total * dVwfm_f2;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vf1cm(2), V_pc, I_pc, "HE");
      Ipc2  := -1.0 * Lookup(V_pc_ref - Vf2cm(2), V_pc, I_pc, "HE");
      Ipu1  := -1.0 * Lookup(V_pu_ref - Vf1cm(2), V_pu, I_pu, "HE");
      Ipu2  := -1.0 * Lookup(V_pu_ref - Vf2cm(2), V_pu, I_pu, "HE");
      Ipd1  :=        Lookup(Vf1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Ipd2  :=        Lookup(Vf2cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vf1cm(2) - V_gc_ref, V_gc, I_gc, "HE");
      Igc2  :=        Lookup(Vf2cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      Term1 := Iout1 + Igc1 - Ipc1;
      Term2 := Ipc2  - Igc2 - Iout2;
      Den   := (Ipu1 * Ipd2) - (Ipd1 * Ipu2);

      -- Since these are falling waveforms
      -- Kf1 == Kpd_on and Kf2 == Kpu_off
      K_tables(Kf1)(i) := (Ipu2*Term1 + Ipu1*Term2) / Den;
      K_tables(Kf2)(i) := (Ipd2*Term1 + Ipd1*Term2) / Den;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " Kr1: " & real'image(K_tables(Kr1)(i)) &
      --       " Kr2: " & real'image(K_tables(Kr2)(i)) &
      --       " Kf1: " & real'image(K_tables(Kf1)(i)) &
      --       " Kf2: " & real'image(K_tables(Kf2)(i));
      -------------------------------------------------------------------------
      -- Shift points by one index for next iteration (for C_comp compensation)
      -------------------------------------------------------------------------
      Vr1cm(1) := Vr1cm(2);
      Vr2cm(1) := Vr2cm(2);
      Vf1cm(1) := Vf1cm(2);
      Vf2cm(1) := Vf2cm(2);

      Vr1cm(2) := Vr1cm(3);
      Vr2cm(2) := Vr2cm(3);
      Vf1cm(2) := Vf1cm(3);
      Vf2cm(2) := Vf2cm(3);

    end loop; -- for loop
    ---------------------------------------------------------------------------
    return K_tables;
  end function Coeff;
--=============================================================================
  constant Ktables : k_tables_type := Coeff(CommonLength, Max_dt,

                                            Ipc_vec, Vpc_vec,
                                            Ipu_vec, Vpu_vec,
                                            Ipd_vec, Vpd_vec,
                                            Igc_vec, Vgc_vec,

                                            Vr1_vec, Tr1_vec,
                                            Vr2_vec, Tr2_vec,
                                            Vf1_vec, Tf1_vec,
                                            Vf2_vec, Tf2_vec,

                                            Vpc_ref_val,
                                            Vpu_ref_val,
                                            Vpd_ref_val,
                                            Vgc_ref_val,

                                            Vfx_r1_val,
                                            Vfx_r2_val,
                                            Vfx_f1_val,
                                            Vfx_f2_val,

                                            Rfx_r1_val,
                                            Rfx_r2_val,
                                            Rfx_f1_val,
                                            Rfx_f2_val,

                                            C_comp_val);
--=============================================================================
begin

  Catch: process is
  -----------------------------------------------------------------------------
  begin
  -----------------------------------------------------------------------------
    if domain = quiescent_domain then

      wait until domain /= quiescent_domain;
      EventState <= 1;

    else

      wait on Vin'above(Vth_R), Vin'above(Vth_F),
              Ven'above(Vth_R), Ven'above(Vth_F);

      if ((Vin > Vth_R) or (Vin < Vth_F)) and
         ((Ven > Vth_R) or (Ven < Vth_F)) then

        if   EventState < 3 then  EventState <= EventState+1;
        else                      EventState <= EventState-1;
        end if;

      end if;
    end if;
  -----------------------------------------------------------------------------
  end process Catch;
  --===========================================================================
  FindState: process (EventState) is
  -----------------------------------------------------------------------------
  begin
  -----------------------------------------------------------------------------
    if (Ven > Vth_R) and (Vin > Vth_R) then           -- high state
      pu_on  <= 1;
      pd_off <= 1;
      pd_on  <= 0;
      pu_off <= 0;
    elsif (Ven > Vth_R) and (Vin < Vth_F) then        -- low state
      pu_on  <= 0;
      pd_off <= 0;
      pd_on  <= 1;
      pu_off <= 1;
    elsif (Ven < Vth_F) then                          -- 3-state
      pu_on  <= 0;
      pd_off <= 1;
      pd_on  <= 0;
      pu_off <= 1;
    else                                              -- unknown state
      pu_on  <= 1;
      pd_off <= 0;
      pd_on  <= 1;
      pu_off <= 0;
    end if;
  -----------------------------------------------------------------------------
  end process FindState;
--=============================================================================
  break on pu_on;
  break on pu_off;
  break on pd_on;
  break on pd_off;
--=============================================================================
  -- This section contains the simultaneous analog equations to find the
  -- appropriate scaling coefficients according to the state the buffer.
  -----------------------------------------------------------------------------
--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
-- This entire section could be made more efficient with nested IF statements
-- if the simulator wouldn't have a problem with simultaneous statements in
-- nested case statements.
--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

  -- Executes during DC operating point calculation
  if    (EventState = 0) and (Ven > Vth_R) and (Vin > Vth_R) use   -- high state
    kpu == Ktables(Kr1)(CommonLength);
  elsif (EventState = 0) and (Ven > Vth_R) and (Vin < Vth_F) use   -- low state
    kpu == Ktables(Kf2)(CommonLength);
  elsif (EventState = 0) and (Ven < Vth_F) use                     -- 3-state
    kpu == Ktables(Kf2)(CommonLength);

  -- Executes after DC operating point calculation before first event
  elsif (EventState = 1) and (pu_off = 1) use
    kpu == Ktables(Kf2)(CommonLength);      -- Last point from pu_off table
  elsif (EventState = 1) and (pu_off /= 1)use
    kpu == Ktables(Kr1)(CommonLength);      -- Last point from pu_on table

  -- Executes after first event (normal operation)
  elsif (EventState > 1) and (pu_on  = 1) use    -- Pullup turning on
    kpu == Lookup(pu_on'last_event,  Ktables(Tcm), Ktables(Kr1), "HE");
  elsif (EventState > 1) and (pu_off = 1) use    -- Pullup turning off
    kpu == Lookup(pu_off'last_event, Ktables(Tcm), Ktables(Kf2), "HE");

  -- Just in case
  else
    kpu == Ktables(Kr1)(Ktables(Kr1)'left); -- 1st point of pullup on table
  end use;


  -- Executes during DC operating point calculation
  if    (EventState = 0) and (Ven > Vth_R) and (Vin > Vth_R) use   -- high state
    kpd == Ktables(Kr2)(CommonLength);
  elsif (EventState = 0) and (Ven > Vth_R) and (Vin < Vth_F) use   -- low state
    kpd == Ktables(Kf1)(CommonLength);
  elsif (EventState = 0) and (Ven < Vth_F) use                     -- 3-state
    kpd == Ktables(Kr2)(CommonLength);

  -- Executes after DC operating point calculation before first event
  elsif (EventState = 1) and (pd_off = 1) use
    kpd == Ktables(Kr2)(CommonLength);      -- Last point from pd_off table
  elsif (EventState = 1) and (pd_off /= 1)use
    kpd == Ktables(Kf1)(CommonLength);      -- Last point from pd_on table

  -- Executes after first event (normal operation)
  elsif (EventState > 1) and (pd_on  = 1) use    -- Pulldown turning on
    kpd == Lookup(pd_on'last_event,  Ktables(Tcm), Ktables(Kf1), "HE");
  elsif (EventState > 1) and (pd_off = 1) use    -- Pulldown turning off
    kpd == Lookup(pd_off'last_event, Ktables(Tcm), Ktables(Kr2), "HE");

  -- Just in case
  else
    kpd == Ktables(Kf1)(Ktables(Kf1)'left); -- 1st point of pulldown on table
  end use;
--=============================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc == -1.0*kI_pc    *Lookup(Vpc, Vpc_vec, Ipc_vec, "SE");-- + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Ipu == -1.0*kI_pu*kpu*Lookup(Vpu, Vpu_vec, Ipu_vec, "SE");-- + kC_comp_pu_val*C_comp_val*Vpu'dot;
  Ipd ==      kI_pd*kpd*Lookup(Vpd, Vpd_vec, Ipd_vec, "SE");-- + kC_comp_pd_val*C_comp_val*Vpd'dot;
  Igc ==      kI_gc    *Lookup(Vgc, Vgc_vec, Igc_vec, "SE");-- + kC_comp_gc_val*C_comp_val*Vgc'dot;
--=============================================================================
end architecture IBIS_2EQ2UK;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_OPENSINK is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.25;       -- splitting coefficients
           kC_comp_pu : real := 0.25;
           kC_comp_pd : real := 0.25;
           kC_comp_gc : real := 0.25;
           --------------------------------------------------------------------
           -- [Pullup Reference] and [Pulldown Reference] values
           --------------------------------------------------------------------
           Vpc_ref : real := 5.0;
           Vpd_ref : real := 0.0;
           Vgc_ref : real := 0.0;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Ipd_data : real_vector := (-0.10,  0.00,  0.10,  0.20);
           Vpd_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Vectors of the Vt curve tables
           --------------------------------------------------------------------
           Vr1_data : real_vector := (2.50,  2.50,     5.00,     5.00);
           Tr1_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           Vf1_data : real_vector := (5.00,  5.00,     2.50,     2.50);
           Tf1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           --------------------------------------------------------------------
           -- V_fixture and R_fixture values
           --------------------------------------------------------------------
           Vfx_r1 : real := 5.0;
           Vfx_f1 : real := 5.0;

           Rfx_r1 : real := 50.0;
           Rfx_f1 : real := 50.0;
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_pd        : real    := 1.0;     -- PD current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling

           kt_rise      : real    := 1.0;     -- Rising waveform time scaling
           kt_fall      : real    := 1.0;     -- Falling waveform time scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           Max_dt       : real    := 1.0e-12; -- Maximum dt between time points
           Vth_R        : real    := 0.8;     -- Input threshold for rising edges
           Vth_F        : real    := 0.2;     -- Input threshold for falling edges
           MaxIV_length : integer := 100;     -- Maximum possible length of IV tables
           MaxVt_length : integer := 1000);   -- Maximum possible length of Vt tables
     --------------------------------------------------------------------------
     port (terminal PU_ref : electrical;
           terminal PD_ref : electrical;

           terminal Pad    : electrical;
           terminal In_D   : electrical;

           terminal PC_ref : electrical;
           terminal GC_ref : electrical);

end entity IBIS_OPENSINK;
--=============================================================================
architecture IBIS_1EQ1UK of IBIS_OPENSINK is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vpu   across  Ipu   through  PU_ref  to  Pad;
  quantity  Vpd   across  Ipd   through  Pad     to  PD_ref;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;

  quantity  Vin   across                 In_D    to  ELECTRICAL_REF;

  signal    pd_on         : integer := 0;
  signal    pd_off        : integer := 0;
  signal    EventState    : integer := 0;

  quantity  kpd           : current := 0.0;
  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_pu_val : real := FileRead(DataFile, "kC_comp_pu", kC_comp_pu);
  constant  kC_comp_pd_val : real := FileRead(DataFile, "kC_comp_pd", kC_comp_pd);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vpc_ref_val : real := FileRead(DataFile, "Vpc_ref", Vpc_ref);
  constant  Vpd_ref_val : real := FileRead(DataFile, "Vpd_ref", Vpd_ref);
  constant  Vgc_ref_val : real := FileRead(DataFile, "Vgc_ref", Vgc_ref);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Ipd_vec     : real_vector := FileRead(DataFile, "Ipd_data", Ipd_data, MaxIV_length);
  constant  Vpd_vec     : real_vector := FileRead(DataFile, "Vpd_data", Vpd_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);

  constant  Vr1_vec     : real_vector := FileRead(DataFile, "Vr1_data", Vr1_data, MaxVt_length);
  constant  Tr1_vec     : real_vector := FileRead(DataFile, "Tr1_data", Tr1_data, MaxVt_length, kt_rise);
  constant  Vf1_vec     : real_vector := FileRead(DataFile, "Vf1_data", Vf1_data, MaxVt_length);
  constant  Tf1_vec     : real_vector := FileRead(DataFile, "Tf1_data", Tf1_data, MaxVt_length, kt_fall);

  constant  Vfx_r1_val  : real := FileRead(DataFile, "Vfx_r1", Vfx_r1);
  constant  Vfx_f1_val  : real := FileRead(DataFile, "Vfx_f1", Vfx_f1);

  constant  Rfx_r1_val  : real := FileRead(DataFile, "Rfx_r1", Rfx_r1);
  constant  Rfx_f1_val  : real := FileRead(DataFile, "Rfx_f1", Rfx_f1);
--=============================================================================
  constant CommonLength : integer := FindCommonLength(Max_dt, Tr1_vec, Tf1_vec);

  type k_elements    is (Tcm, Kr1, Kf1);
  type k_tables_type is array(k_elements) of real_vector(1 to CommonLength);
--=============================================================================
  function Coeff (Common_length : integer;
                  Max_dt : real;
                  I_pc   : real_vector;
                  V_pc   : real_vector;
                  I_pd   : real_vector;
                  V_pd   : real_vector;
                  I_gc   : real_vector;
                  V_gc   : real_vector;

                  V_r1   : real_vector;
                  T_r1   : real_vector;
                  V_f1   : real_vector;
                  T_f1   : real_vector;

                  V_pc_ref : real;
                  V_pd_ref : real;
                  V_gc_ref : real;

                  V_fx_r1  : real;
                  V_fx_f1  : real;

                  R_fx_r1  : real;
                  R_fx_f1  : real;

                  C_comp_total : real) return k_tables_type is
  -----------------------------------------------------------------------------
  -- This function converts each Vt table to a corresponding scaling
  -- coefficient table that is used to scale the IV curves with respect
  -- to time.  The C_comp effects in the waveforms are eliminated from the
  -- scaling coefficients, so that the output waveforms will match the Vt
  -- table regardless of the presence of C_comp under identical loading
  -- conditions, i.e. C_comp in the output stage will not derate the
  -- original Vt curves.
  -----------------------------------------------------------------------------
    variable K_tables : k_tables_type;

    variable Vr1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf1cm : real_vector(1 to 3) := (others => 0.0);

    variable dVwfm_r1 : real := 0.0;
    variable dVwfm_f1 : real := 0.0;

    variable Iout1 : real := 0.0;
    variable Ipc1  : real := 0.0;
    variable Ipd1  : real := 0.0;
    variable Igc1  : real := 0.0;
  -----------------------------------------------------------------------------
  begin
    K_tables(Tcm) := CommonTime(Common_length, Max_dt, T_r1, T_f1);
    --=========================================================================
    -- Calculate the scaling coefficients for each time point along Tcm
    ---------------------------------------------------------------------------
    -- Store the first voltage point in the common waveform variables
    ---------------------------------------------------------------------------
    Vr1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r1, V_r1, "HE");
    Vf1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f1, V_f1, "HE");

    for i in K_tables(Tcm)'range loop
    --report "Tcm: " & real'image(K_tables(Tcm)(i));  -- For debugging
    ---------------------------------------------------------------------------
    -- Store the next point (for the derivative calculations)
    ---------------------------------------------------------------------------
      if (i < K_tables(Tcm)'right) then
        Vr1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r1, V_r1, "HE");
        Vf1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f1, V_f1, "HE");
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: " & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " R1: " & real'image(Vr1cm(2)) &
      --       " F1: " & real'image(Vf1cm(2));
      -------------------------------------------------------------------------
      -- Calculate the derivative of each waveform for C_comp compensation
      -------------------------------------------------------------------------
      if ((i <= K_tables(Tcm)'left) or (i >= K_tables(Tcm)'right)) then
        dVwfm_r1 := 0.0;          -- Force the end point derivatives to zero
        dVwfm_f1 := 0.0;
      else
        dVwfm_r1 := ((Vr1cm(2) - Vr1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr1cm(3) - Vr1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f1 := ((Vf1cm(2) - Vf1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf1cm(3) - Vf1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " dR1: " & real'image(dVwfm_r1) &
      --       " dF1: " & real'image(dVwfm_f1);
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kr1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vr1cm(2) - V_fx_r1) / R_fx_r1) + C_comp_total * dVwfm_r1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vr1cm(2), V_pc, I_pc, "HE");
      Ipd1  :=        Lookup(Vr1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vr1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are rising waveforms
      -- Kr1 == Kpd_off
      if (Ipd1 = 0.0) then K_tables(Kr1)(i) := 0.0;
      else                 K_tables(Kr1)(i) := (Ipc1 - Igc1 - Iout1) / Ipd1;
      end if;
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kf1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vf1cm(2) - V_fx_f1) / R_fx_f1) + C_comp_total * dVwfm_f1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vf1cm(2), V_pc, I_pc, "HE");
      Ipd1  :=        Lookup(Vf1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vf1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are falling waveforms
      -- Kf1 == Kpd_on
      if (Ipd1 = 0.0) then K_tables(Kf1)(i) := 0.0;
      else                 K_tables(Kf1)(i) := (Ipc1 - Igc1 - Iout1) / Ipd1;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " Kr1: " & real'image(K_tables(Kr1)(i)) &
      --       " Kf1: " & real'image(K_tables(Kf1)(i));
      -------------------------------------------------------------------------
      -- Shift points by one index for next iteration (for C_comp compensation)
      -------------------------------------------------------------------------
      Vr1cm(1) := Vr1cm(2);
      Vf1cm(1) := Vf1cm(2);

      Vr1cm(2) := Vr1cm(3);
      Vf1cm(2) := Vf1cm(3);

    end loop; -- for loop
    ---------------------------------------------------------------------------
    return K_tables;
  end function Coeff;
--=============================================================================
  constant Ktables : k_tables_type := Coeff(CommonLength, Max_dt,

                                            Ipc_vec, Vpc_vec,
                                            Ipd_vec, Vpd_vec,
                                            Igc_vec, Vgc_vec,

                                            Vr1_vec, Tr1_vec,
                                            Vf1_vec, Tf1_vec,

                                            Vpc_ref_val,
                                            Vpd_ref_val,
                                            Vgc_ref_val,

                                            Vfx_r1_val,
                                            Vfx_f1_val,

                                            Rfx_r1_val,
                                            Rfx_f1_val,

                                            C_comp_val);
--=============================================================================
begin

  Catch: process is
  -----------------------------------------------------------------------------
  begin
  -----------------------------------------------------------------------------
    if domain = quiescent_domain then

      wait until domain /= quiescent_domain;
      EventState <= 1;

    else

      wait on Vin'above(Vth_R), Vin'above(Vth_F);

      if ((Vin > Vth_R) or (Vin < Vth_F)) then

        if   EventState < 3 then  EventState <= EventState+1;
        else                      EventState <= EventState-1;
        end if;

      end if;
    end if;
  -----------------------------------------------------------------------------
  end process Catch;
  --===========================================================================
  FindState: process (EventState) is
  -----------------------------------------------------------------------------
  begin
  -----------------------------------------------------------------------------
    if (Vin > Vth_R) then                             -- high state
      pd_off <= 1;
      pd_on  <= 0;
    elsif (Vin < Vth_F) then                          -- low state
      pd_off <= 0;
      pd_on  <= 1;
    else                                              -- unknown state
      pd_off <= 0;
      pd_on  <= 1;
    end if;
  -----------------------------------------------------------------------------
  end process FindState;
--=============================================================================
  break on pd_on;
  break on pd_off;
--=============================================================================
  -- This section contains the simultaneous analog equations to find the
  -- appropriate scaling coefficients according to the state the buffer.
  -----------------------------------------------------------------------------
--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
-- This entire section could be made more efficient with nested IF statements
-- if the simulator wouldn't have a problem with simultaneous statements in
-- nested case statements.
--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

  -- Executes during DC operating point calculation
  if    (EventState = 0) and (Vin > Vth_R) use                     -- high state
    kpd == Ktables(Kr1)(CommonLength);
  elsif (EventState = 0) and (Vin < Vth_F) use                     -- low state
    kpd == Ktables(Kf1)(CommonLength);

  -- Executes after DC operating point calculation before first event
  elsif (EventState = 1) and (pd_off = 1) use
    kpd == Ktables(Kr1)(CommonLength);      -- Last point from pd_off table
  elsif (EventState = 1) and (pd_off /= 1)use
    kpd == Ktables(Kf1)(CommonLength);      -- Last point from pd_on table

  -- Executes after first event (normal operation)
  elsif (EventState > 1) and (pd_on  = 1) use    -- Pulldown turning on
    kpd == Lookup(pd_on'last_event,  Ktables(Tcm), Ktables(Kf1), "HE");
  elsif (EventState > 1) and (pd_off = 1) use    -- Pulldown turning off
    kpd == Lookup(pd_off'last_event, Ktables(Tcm), Ktables(Kr1), "HE");

  -- Just in case
  else
    kpd == Ktables(Kf1)(Ktables(Kf1)'left); -- 1st point of pulldown on table
  end use;
--=============================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc == -1.0*kI_pc    *Lookup(Vpc, Vpc_vec, Ipc_vec, "SE");-- + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Ipu ==  0.0                                              ;-- + kC_comp_pu_val*C_comp_val*Vpu'dot;
  Ipd ==      kI_pd*kpd*Lookup(Vpd, Vpd_vec, Ipd_vec, "SE");-- + kC_comp_pd_val*C_comp_val*Vpd'dot;
  Igc ==      kI_gc    *Lookup(Vgc, Vgc_vec, Igc_vec, "SE");-- + kC_comp_gc_val*C_comp_val*Vgc'dot;
--=============================================================================
end architecture IBIS_1EQ1UK;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_IO_OPENSINK is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.25;       -- splitting coefficients
           kC_comp_pu : real := 0.25;
           kC_comp_pd : real := 0.25;
           kC_comp_gc : real := 0.25;
           --------------------------------------------------------------------
           -- Receiver thresholds
           --------------------------------------------------------------------
           Vinh : real := 2.00;
           Vinl : real := 0.80;
           --------------------------------------------------------------------
           -- [Pullup Reference] and [Pulldown Reference] values
           --------------------------------------------------------------------
           Vpc_ref : real := 5.0;
           Vpd_ref : real := 0.0;
           Vgc_ref : real := 0.0;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Ipd_data : real_vector := (-0.10,  0.00,  0.10,  0.20);
           Vpd_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Vectors of the Vt curve tables
           --------------------------------------------------------------------
           Vr1_data : real_vector := (2.50,  2.50,     5.00,     5.00);
           Tr1_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           Vf1_data : real_vector := (5.00,  5.00,     2.50,     2.50);
           Tf1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           --------------------------------------------------------------------
           -- V_fixture and R_fixture values
           --------------------------------------------------------------------
           Vfx_r1 : real := 5.0;
           Vfx_f1 : real := 5.0;

           Rfx_r1 : real := 50.0;
           Rfx_f1 : real := 50.0;
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_pd        : real    := 1.0;     -- PD current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling

           kt_rise      : real    := 1.0;     -- Rising waveform time scaling
           kt_fall      : real    := 1.0;     -- Falling waveform time scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           Max_dt       : real    := 1.0e-12; -- Maximum dt between time points
           Vth_R        : real    := 0.8;     -- Input threshold for rising edges
           Vth_F        : real    := 0.2;     -- Input threshold for falling edges
           MaxIV_length : integer := 100;     -- Maximum possible length of IV tables
           MaxVt_length : integer := 1000);   -- Maximum possible length of Vt tables
     --------------------------------------------------------------------------
     port (terminal PU_ref : electrical;
           terminal PD_ref : electrical;

           terminal Pad    : electrical;
           terminal In_D   : electrical;
           terminal En_D   : electrical;
--           terminal Rcv_D  : electrical;

           terminal PC_ref : electrical;
           terminal GC_ref : electrical);

end entity IBIS_IO_OPENSINK;
--=============================================================================
architecture IBIS_1EQ1UK of IBIS_IO_OPENSINK is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vpu   across  Ipu   through  PU_ref  to  Pad;
  quantity  Vpd   across  Ipd   through  Pad     to  PD_ref;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;

  quantity  Vin   across                 In_D    to  ELECTRICAL_REF;
  quantity  Ven   across                 En_D    to  ELECTRICAL_REF;
--  quantity  Vrcv  across  Ircv  through  Rcv_D   to  ELECTRICAL_REF;

  signal    pd_on         : integer := 0;
  signal    pd_off        : integer := 0;
  signal    EventState    : integer := 0;

  quantity  kpd           : current := 0.0;
  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_pu_val : real := FileRead(DataFile, "kC_comp_pu", kC_comp_pu);
  constant  kC_comp_pd_val : real := FileRead(DataFile, "kC_comp_pd", kC_comp_pd);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vinh_val    : real := FileRead(DataFile, "Vinh", Vinh);
  constant  Vinl_val    : real := FileRead(DataFile, "Vinl", Vinl);

  constant  Vpc_ref_val : real := FileRead(DataFile, "Vpc_ref", Vpc_ref);
  constant  Vpd_ref_val : real := FileRead(DataFile, "Vpd_ref", Vpd_ref);
  constant  Vgc_ref_val : real := FileRead(DataFile, "Vgc_ref", Vgc_ref);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Ipd_vec     : real_vector := FileRead(DataFile, "Ipd_data", Ipd_data, MaxIV_length);
  constant  Vpd_vec     : real_vector := FileRead(DataFile, "Vpd_data", Vpd_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);

  constant  Vr1_vec     : real_vector := FileRead(DataFile, "Vr1_data", Vr1_data, MaxVt_length);
  constant  Tr1_vec     : real_vector := FileRead(DataFile, "Tr1_data", Tr1_data, MaxVt_length, kt_rise);
  constant  Vf1_vec     : real_vector := FileRead(DataFile, "Vf1_data", Vf1_data, MaxVt_length);
  constant  Tf1_vec     : real_vector := FileRead(DataFile, "Tf1_data", Tf1_data, MaxVt_length, kt_fall);

  constant  Vfx_r1_val  : real := FileRead(DataFile, "Vfx_r1", Vfx_r1);
  constant  Vfx_f1_val  : real := FileRead(DataFile, "Vfx_f1", Vfx_f1);

  constant  Rfx_r1_val  : real := FileRead(DataFile, "Rfx_r1", Rfx_r1);
  constant  Rfx_f1_val  : real := FileRead(DataFile, "Rfx_f1", Rfx_f1);
--=============================================================================
  constant CommonLength : integer := FindCommonLength(Max_dt, Tr1_vec, Tf1_vec);

  type k_elements    is (Tcm, Kr1, Kf1);
  type k_tables_type is array(k_elements) of real_vector(1 to CommonLength);
--=============================================================================
  function Coeff (Common_length : integer;
                  Max_dt : real;
                  I_pc   : real_vector;
                  V_pc   : real_vector;
                  I_pd   : real_vector;
                  V_pd   : real_vector;
                  I_gc   : real_vector;
                  V_gc   : real_vector;

                  V_r1   : real_vector;
                  T_r1   : real_vector;
                  V_f1   : real_vector;
                  T_f1   : real_vector;

                  V_pc_ref : real;
                  V_pd_ref : real;
                  V_gc_ref : real;

                  V_fx_r1  : real;
                  V_fx_f1  : real;

                  R_fx_r1  : real;
                  R_fx_f1  : real;

                  C_comp_total : real) return k_tables_type is
  -----------------------------------------------------------------------------
  -- This function converts each Vt table to a corresponding scaling
  -- coefficient table that is used to scale the IV curves with respect
  -- to time.  The C_comp effects in the waveforms are eliminated from the
  -- scaling coefficients, so that the output waveforms will match the Vt
  -- table regardless of the presence of C_comp under identical loading
  -- conditions, i.e. C_comp in the output stage will not derate the
  -- original Vt curves.
  -----------------------------------------------------------------------------
    variable K_tables : k_tables_type;

    variable Vr1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf1cm : real_vector(1 to 3) := (others => 0.0);

    variable dVwfm_r1 : real := 0.0;
    variable dVwfm_f1 : real := 0.0;

    variable Iout1 : real := 0.0;
    variable Ipc1  : real := 0.0;
    variable Ipd1  : real := 0.0;
    variable Igc1  : real := 0.0;
  -----------------------------------------------------------------------------
  begin
    K_tables(Tcm) := CommonTime(Common_length, Max_dt, T_r1, T_f1);
    --=========================================================================
    -- Calculate the scaling coefficients for each time point along Tcm
    ---------------------------------------------------------------------------
    -- Store the first voltage point in the common waveform variables
    ---------------------------------------------------------------------------
    Vr1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r1, V_r1, "HE");
    Vf1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f1, V_f1, "HE");

    for i in K_tables(Tcm)'range loop
    --report "Tcm: " & real'image(K_tables(Tcm)(i));  -- For debugging
    ---------------------------------------------------------------------------
    -- Store the next point (for the derivative calculations)
    ---------------------------------------------------------------------------
      if (i < K_tables(Tcm)'right) then
        Vr1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r1, V_r1, "HE");
        Vf1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f1, V_f1, "HE");
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: " & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " R1: " & real'image(Vr1cm(2)) &
      --       " F1: " & real'image(Vf1cm(2));
      -------------------------------------------------------------------------
      -- Calculate the derivative of each waveform for C_comp compensation
      -------------------------------------------------------------------------
      if ((i <= K_tables(Tcm)'left) or (i >= K_tables(Tcm)'right)) then
        dVwfm_r1 := 0.0;          -- Force the end point derivatives to zero
        dVwfm_f1 := 0.0;
      else
        dVwfm_r1 := ((Vr1cm(2) - Vr1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr1cm(3) - Vr1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f1 := ((Vf1cm(2) - Vf1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf1cm(3) - Vf1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " dR1: " & real'image(dVwfm_r1) &
      --       " dF1: " & real'image(dVwfm_f1);
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kr1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vr1cm(2) - V_fx_r1) / R_fx_r1) + C_comp_total * dVwfm_r1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vr1cm(2), V_pc, I_pc, "HE");
      Ipd1  :=        Lookup(Vr1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vr1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are rising waveforms
      -- Kr1 == Kpd_off
      if (Ipd1 = 0.0) then K_tables(Kr1)(i) := 0.0;
      else                 K_tables(Kr1)(i) := (Ipc1 - Igc1 - Iout1) / Ipd1;
      end if;
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kf1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vf1cm(2) - V_fx_f1) / R_fx_f1) + C_comp_total * dVwfm_f1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vf1cm(2), V_pc, I_pc, "HE");
      Ipd1  :=        Lookup(Vf1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vf1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are falling waveforms
      -- Kf1 == Kpd_on
      if (Ipd1 = 0.0) then K_tables(Kf1)(i) := 0.0;
      else                 K_tables(Kf1)(i) := (Ipc1 - Igc1 - Iout1) / Ipd1;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " Kr1: " & real'image(K_tables(Kr1)(i)) &
      --       " Kf1: " & real'image(K_tables(Kf1)(i));
      -------------------------------------------------------------------------
      -- Shift points by one index for next iteration (for C_comp compensation)
      -------------------------------------------------------------------------
      Vr1cm(1) := Vr1cm(2);
      Vf1cm(1) := Vf1cm(2);

      Vr1cm(2) := Vr1cm(3);
      Vf1cm(2) := Vf1cm(3);

    end loop; -- for loop
    ---------------------------------------------------------------------------
    return K_tables;
  end function Coeff;
--=============================================================================
  constant Ktables : k_tables_type := Coeff(CommonLength, Max_dt,

                                            Ipc_vec, Vpc_vec,
                                            Ipd_vec, Vpd_vec,
                                            Igc_vec, Vgc_vec,

                                            Vr1_vec, Tr1_vec,
                                            Vf1_vec, Tf1_vec,

                                            Vpc_ref_val,
                                            Vpd_ref_val,
                                            Vgc_ref_val,

                                            Vfx_r1_val,
                                            Vfx_f1_val,

                                            Rfx_r1_val,
                                            Rfx_f1_val,

                                            C_comp_val);
--=============================================================================
begin

  Catch: process is
  -----------------------------------------------------------------------------
  begin
  -----------------------------------------------------------------------------
    if domain = quiescent_domain then

      wait until domain /= quiescent_domain;
      EventState <= 1;

    else

      wait on Vin'above(Vth_R), Vin'above(Vth_F),
              Ven'above(Vth_R), Ven'above(Vth_F);

      if ((Vin > Vth_R) or (Vin < Vth_F)) and
         ((Ven > Vth_R) or (Ven < Vth_F)) then

        if   EventState < 3 then  EventState <= EventState+1;
        else                      EventState <= EventState-1;
        end if;

      end if;
    end if;
  -----------------------------------------------------------------------------
  end process Catch;
  --===========================================================================
  FindState: process (EventState) is
  -----------------------------------------------------------------------------
  begin
  -----------------------------------------------------------------------------
    if (Ven > Vth_R) and (Vin > Vth_R) then           -- high state
      pd_off <= 1;
      pd_on  <= 0;
    elsif (Ven > Vth_R) and (Vin < Vth_F) then        -- low state
      pd_off <= 0;
      pd_on  <= 1;
    elsif (Ven < Vth_F) then                          -- 3-state
      pd_off <= 1;
      pd_on  <= 0;
    else                                              -- unknown state
      pd_off <= 0;
      pd_on  <= 1;
    end if;
  -----------------------------------------------------------------------------
  end process FindState;
--=============================================================================
  break on pd_on;
  break on pd_off;
--=============================================================================
  -- This section contains the simultaneous analog equations to find the
  -- appropriate scaling coefficients according to the state the buffer.
  -----------------------------------------------------------------------------
--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
-- This entire section could be made more efficient with nested IF statements
-- if the simulator wouldn't have a problem with simultaneous statements in
-- nested case statements.
--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

  -- Executes during DC operating point calculation
  if    (EventState = 0) and (Ven > Vth_R) and (Vin > Vth_R) use   -- high state
    kpd == Ktables(Kr1)(CommonLength);
  elsif (EventState = 0) and (Ven > Vth_R) and (Vin < Vth_F) use   -- low state
    kpd == Ktables(Kf1)(CommonLength);
  elsif (EventState = 0) and (Ven < Vth_F) use                     -- 3-state
    kpd == Ktables(Kr1)(CommonLength);

  -- Executes after DC operating point calculation before first event
  elsif (EventState = 1) and (pd_off = 1) use
    kpd == Ktables(Kr1)(CommonLength);      -- Last point from pd_off table
  elsif (EventState = 1) and (pd_off /= 1)use
    kpd == Ktables(Kf1)(CommonLength);      -- Last point from pd_on table

  -- Executes after first event (normal operation)
  elsif (EventState > 1) and (pd_on  = 1) use    -- Pulldown turning on
    kpd == Lookup(pd_on'last_event,  Ktables(Tcm), Ktables(Kf1), "HE");
  elsif (EventState > 1) and (pd_off = 1) use    -- Pulldown turning off
    kpd == Lookup(pd_off'last_event, Ktables(Tcm), Ktables(Kr1), "HE");

  -- Just in case
  else
    kpd == Ktables(Kf1)(Ktables(Kf1)'left); -- 1st point of pulldown on table
  end use;
--=============================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc == -1.0*kI_pc    *Lookup(Vpc, Vpc_vec, Ipc_vec, "SE");-- + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Ipu ==  0.0                                              ;-- + kC_comp_pu_val*C_comp_val*Vpu'dot;
  Ipd ==      kI_pd*kpd*Lookup(Vpd, Vpd_vec, Ipd_vec, "SE");-- + kC_comp_pd_val*C_comp_val*Vpd'dot;
  Igc ==      kI_gc    *Lookup(Vgc, Vgc_vec, Igc_vec, "SE");-- + kC_comp_gc_val*C_comp_val*Vgc'dot;
  -----------------------------------------------------------------------------
  -- A simple receiver logic
  -----------------------------------------------------------------------------
--  if    (Vpd > Vinh_val) use Vrcv == 1.0;
--  elsif (Vpd < Vinl_val) use Vrcv == 0.0;
--  else                       Vrcv == 0.5;
--  end use;
--=============================================================================
end architecture IBIS_1EQ1UK;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_OPENSOURCE is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.25;       -- splitting coefficients
           kC_comp_pu : real := 0.25;
           kC_comp_pd : real := 0.25;
           kC_comp_gc : real := 0.25;
           --------------------------------------------------------------------
           -- [Pullup Reference] and [Pulldown Reference] values
           --------------------------------------------------------------------
           Vpc_ref : real := 5.0;
           Vpu_ref : real := 5.0;
           Vgc_ref : real := 0.0;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Ipu_data : real_vector := ( 0.10,  0.00, -0.10, -0.20);
           Vpu_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Vectors of the Vt curve tables
           --------------------------------------------------------------------
           Vr1_data : real_vector := (0.00,  0.00,     2.50,     2.50);
           Tr1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vf1_data : real_vector := (2.50,  2.50,     0.00,     0.00);
           Tf1_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           --------------------------------------------------------------------
           -- V_fixture and R_fixture values
           --------------------------------------------------------------------
           Vfx_r1 : real := 0.0;
           Vfx_f1 : real := 0.0;

           Rfx_r1 : real := 50.0;
           Rfx_f1 : real := 50.0;
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_pu        : real    := 1.0;     -- PU current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling

           kt_rise      : real    := 1.0;     -- Rising waveform time scaling
           kt_fall      : real    := 1.0;     -- Falling waveform time scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           Max_dt       : real    := 1.0e-12; -- Maximum dt between time points
           Vth_R        : real    := 0.8;     -- Input threshold for rising edges
           Vth_F        : real    := 0.2;     -- Input threshold for falling edges
           MaxIV_length : integer := 100;     -- Maximum possible length of IV tables
           MaxVt_length : integer := 1000);   -- Maximum possible length of Vt tables
     --------------------------------------------------------------------------
     port (terminal PU_ref : electrical;
           terminal PD_ref : electrical;

           terminal Pad    : electrical;
           terminal In_D   : electrical;

           terminal PC_ref : electrical;
           terminal GC_ref : electrical);

end entity IBIS_OPENSOURCE;
--=============================================================================
architecture IBIS_1EQ1UK of IBIS_OPENSOURCE is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vpu   across  Ipu   through  PU_ref  to  Pad;
  quantity  Vpd   across  Ipd   through  Pad     to  PD_ref;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;

  quantity  Vin   across                 In_D    to  ELECTRICAL_REF;

  signal    pu_on         : integer := 0;
  signal    pu_off        : integer := 0;
  signal    EventState    : integer := 0;

  quantity  kpu           : current := 0.0;
  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_pu_val : real := FileRead(DataFile, "kC_comp_pu", kC_comp_pu);
  constant  kC_comp_pd_val : real := FileRead(DataFile, "kC_comp_pd", kC_comp_pd);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vpc_ref_val : real := FileRead(DataFile, "Vpc_ref", Vpc_ref);
  constant  Vpu_ref_val : real := FileRead(DataFile, "Vpu_ref", Vpu_ref);
  constant  Vgc_ref_val : real := FileRead(DataFile, "Vgc_ref", Vgc_ref);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Ipu_vec     : real_vector := FileRead(DataFile, "Ipu_data", Ipu_data, MaxIV_length);
  constant  Vpu_vec     : real_vector := FileRead(DataFile, "Vpu_data", Vpu_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);

  constant  Vr1_vec     : real_vector := FileRead(DataFile, "Vr1_data", Vr1_data, MaxVt_length);
  constant  Tr1_vec     : real_vector := FileRead(DataFile, "Tr1_data", Tr1_data, MaxVt_length, kt_rise);
  constant  Vf1_vec     : real_vector := FileRead(DataFile, "Vf1_data", Vf1_data, MaxVt_length);
  constant  Tf1_vec     : real_vector := FileRead(DataFile, "Tf1_data", Tf1_data, MaxVt_length, kt_fall);

  constant  Vfx_r1_val  : real := FileRead(DataFile, "Vfx_r1", Vfx_r1);
  constant  Vfx_f1_val  : real := FileRead(DataFile, "Vfx_f1", Vfx_f1);

  constant  Rfx_r1_val  : real := FileRead(DataFile, "Rfx_r1", Rfx_r1);
  constant  Rfx_f1_val  : real := FileRead(DataFile, "Rfx_f1", Rfx_f1);
--=============================================================================
  constant CommonLength : integer := FindCommonLength(Max_dt, Tr1_vec, Tf1_vec);

  type k_elements    is (Tcm, Kr1, Kf1);
  type k_tables_type is array(k_elements) of real_vector(1 to CommonLength);
--=============================================================================
  function Coeff (Common_length : integer;
                  Max_dt : real;
                  I_pc   : real_vector;
                  V_pc   : real_vector;
                  I_pu   : real_vector;
                  V_pu   : real_vector;
                  I_gc   : real_vector;
                  V_gc   : real_vector;

                  V_r1   : real_vector;
                  T_r1   : real_vector;
                  V_f1   : real_vector;
                  T_f1   : real_vector;

                  V_pc_ref : real;
                  V_pu_ref : real;
                  V_gc_ref : real;

                  V_fx_r1  : real;
                  V_fx_f1  : real;

                  R_fx_r1  : real;
                  R_fx_f1  : real;

                  C_comp_total : real) return k_tables_type is
  -----------------------------------------------------------------------------
  -- This function converts each Vt table to a corresponding scaling
  -- coefficient table that is used to scale the IV curves with respect
  -- to time.  The C_comp effects in the waveforms are eliminated from the
  -- scaling coefficients, so that the output waveforms will match the Vt
  -- table regardless of the presence of C_comp under identical loading
  -- conditions, i.e. C_comp in the output stage will not derate the
  -- original Vt curves.
  -----------------------------------------------------------------------------
    variable K_tables : k_tables_type;

    variable Vr1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf1cm : real_vector(1 to 3) := (others => 0.0);

    variable dVwfm_r1 : real := 0.0;
    variable dVwfm_f1 : real := 0.0;

    variable Iout1 : real := 0.0;
    variable Ipc1  : real := 0.0;
    variable Ipu1  : real := 0.0;
    variable Ipd1  : real := 0.0;
    variable Igc1  : real := 0.0;
  -----------------------------------------------------------------------------
  begin
    K_tables(Tcm) := CommonTime(Common_length, Max_dt, T_r1, T_f1);
    --=========================================================================
    -- Calculate the scaling coefficients for each time point along Tcm
    ---------------------------------------------------------------------------
    -- Store the first voltage point in the common waveform variables
    ---------------------------------------------------------------------------
    Vr1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r1, V_r1, "HE");
    Vf1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f1, V_f1, "HE");

    for i in K_tables(Tcm)'range loop
    --report "Tcm: " & real'image(K_tables(Tcm)(i));  -- For debugging
    ---------------------------------------------------------------------------
    -- Store the next point (for the derivative calculations)
    ---------------------------------------------------------------------------
      if (i < K_tables(Tcm)'right) then
        Vr1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r1, V_r1, "HE");
        Vf1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f1, V_f1, "HE");
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: " & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " R1: " & real'image(Vr1cm(2)) &
      --       " F1: " & real'image(Vf1cm(2));
      -------------------------------------------------------------------------
      -- Calculate the derivative of each waveform for C_comp compensation
      -------------------------------------------------------------------------
      if ((i <= K_tables(Tcm)'left) or (i >= K_tables(Tcm)'right)) then
        dVwfm_r1 := 0.0;          -- Force the end point derivatives to zero
        dVwfm_f1 := 0.0;
      else
        dVwfm_r1 := ((Vr1cm(2) - Vr1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr1cm(3) - Vr1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f1 := ((Vf1cm(2) - Vf1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf1cm(3) - Vf1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " dR1: " & real'image(dVwfm_r1) &
      --       " dF1: " & real'image(dVwfm_f1);
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kr1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vr1cm(2) - V_fx_r1) / R_fx_r1) + C_comp_total * dVwfm_r1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vr1cm(2), V_pc, I_pc, "HE");
      Ipu1  :=        Lookup(V_pu_ref - Vr1cm(2), V_pu, I_pu, "HE");
      Igc1  :=        Lookup(Vr1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are rising waveforms
      -- Kr1 == Kpu_on
      if (Ipu1 = 0.0) then K_tables(Kr1)(i) := 0.0;
      else                 K_tables(Kr1)(i) := (Ipc1 - Igc1 - Iout1) / Ipu1;
      end if;
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kf1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vf1cm(2) - V_fx_f1) / R_fx_f1) + C_comp_total * dVwfm_f1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vf1cm(2), V_pc, I_pc, "HE");
      Ipu1  :=        Lookup(V_pu_ref - Vf1cm(2), V_pu, I_pu, "HE");
      Igc1  :=        Lookup(Vf1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are falling waveforms
      -- Kf1 == Kpu_off
      if (Ipu1 = 0.0) then K_tables(Kf1)(i) := 0.0;
      else                 K_tables(Kf1)(i) := (Ipc1 - Igc1 - Iout1) / Ipu1;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " Kr1: " & real'image(K_tables(Kr1)(i)) &
      --       " Kf1: " & real'image(K_tables(Kf1)(i));
      -------------------------------------------------------------------------
      -- Shift points by one index for next iteration (for C_comp compensation)
      -------------------------------------------------------------------------
      Vr1cm(1) := Vr1cm(2);
      Vf1cm(1) := Vf1cm(2);

      Vr1cm(2) := Vr1cm(3);
      Vf1cm(2) := Vf1cm(3);

    end loop; -- for loop
    ---------------------------------------------------------------------------
    return K_tables;
  end function Coeff;
--=============================================================================
  constant Ktables : k_tables_type := Coeff(CommonLength, Max_dt,

                                            Ipc_vec, Vpc_vec,
                                            Ipu_vec, Vpu_vec,
                                            Igc_vec, Vgc_vec,

                                            Vr1_vec, Tr1_vec,
                                            Vf1_vec, Tf1_vec,

                                            Vpc_ref_val,
                                            Vpu_ref_val,
                                            Vgc_ref_val,

                                            Vfx_r1_val,
                                            Vfx_f1_val,

                                            Rfx_r1_val,
                                            Rfx_f1_val,

                                            C_comp_val);
--=============================================================================
begin

  Catch: process is
  -----------------------------------------------------------------------------
  begin
  -----------------------------------------------------------------------------
    if domain = quiescent_domain then

      wait until domain /= quiescent_domain;
      EventState <= 1;

    else

      wait on Vin'above(Vth_R), Vin'above(Vth_F);

      if ((Vin > Vth_R) or (Vin < Vth_F)) then

        if   EventState < 3 then  EventState <= EventState+1;
        else                      EventState <= EventState-1;
        end if;

      end if;
    end if;
  -----------------------------------------------------------------------------
  end process Catch;
  --===========================================================================
  FindState: process (EventState) is
  -----------------------------------------------------------------------------
  begin
  -----------------------------------------------------------------------------
    if (Vin > Vth_R) then                             -- high state
      pu_on  <= 1;
      pu_off <= 0;
    elsif (Vin < Vth_F) then                          -- low state
      pu_on  <= 0;
      pu_off <= 1;
    else                                              -- unknown state
      pu_on  <= 1;
      pu_off <= 0;
    end if;
  -----------------------------------------------------------------------------
  end process FindState;
--=============================================================================
  break on pu_on;
  break on pu_off;
--=============================================================================
  -- This section contains the simultaneous analog equations to find the
  -- appropriate scaling coefficients according to the state the buffer.
  -----------------------------------------------------------------------------
--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
-- This entire section could be made more efficient with nested IF statements
-- if the simulator wouldn't have a problem with simultaneous statements in
-- nested case statements.
--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

  -- Executes during DC operating point calculation
  if    (EventState = 0) and (Vin > Vth_R) use                     -- high state
    kpu == Ktables(Kr1)(CommonLength);
  elsif (EventState = 0) and (Vin < Vth_F) use                     -- low state
    kpu == Ktables(Kf1)(CommonLength);

  -- Executes after DC operating point calculation before first event
  elsif (EventState = 1) and (pu_off = 1) use
    kpu == Ktables(Kf1)(CommonLength);      -- Last point from pu_off table
  elsif (EventState = 1) and (pu_off /= 1)use
    kpu == Ktables(Kr1)(CommonLength);      -- Last point from pu_on table

  -- Executes after first event (normal operation)
  elsif (EventState > 1) and (pu_on  = 1) use    -- Pullup turning on
    kpu == Lookup(pu_on'last_event,  Ktables(Tcm), Ktables(Kr1), "HE");
  elsif (EventState > 1) and (pu_off = 1) use    -- Pullup turning off
    kpu == Lookup(pu_off'last_event, Ktables(Tcm), Ktables(Kf1), "HE");

  -- Just in case
  else
    kpu == Ktables(Kr1)(Ktables(Kr1)'left); -- 1st point of pullup on table
  end use;

--=============================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc == -1.0*kI_pc    *Lookup(Vpc, Vpc_vec, Ipc_vec, "SE");-- + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Ipu == -1.0*kI_pu*kpu*Lookup(Vpu, Vpu_vec, Ipu_vec, "SE");-- + kC_comp_pu_val*C_comp_val*Vpu'dot;
  Ipd ==  0.0                                              ;-- + kC_comp_pd_val*C_comp_val*Vpd'dot;
  Igc ==      kI_gc    *Lookup(Vgc, Vgc_vec, Igc_vec, "SE");-- + kC_comp_gc_val*C_comp_val*Vgc'dot;
--=============================================================================
end architecture IBIS_1EQ1UK;
--=============================================================================
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_IO_OPENSOURCE is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.25;       -- splitting coefficients
           kC_comp_pu : real := 0.25;
           kC_comp_pd : real := 0.25;
           kC_comp_gc : real := 0.25;
           --------------------------------------------------------------------
           -- Receiver thresholds
           --------------------------------------------------------------------
           Vinh : real := 2.00;
           Vinl : real := 0.80;
           --------------------------------------------------------------------
           -- [Pullup Reference] and [Pulldown Reference] values
           --------------------------------------------------------------------
           Vpc_ref : real := 5.0;
           Vpu_ref : real := 5.0;
           Vgc_ref : real := 0.0;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Ipu_data : real_vector := ( 0.10,  0.00, -0.10, -0.20);
           Vpu_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Vectors of the Vt curve tables
           --------------------------------------------------------------------
           Vr1_data : real_vector := (0.00,  0.00,     2.50,     2.50);
           Tr1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vf1_data : real_vector := (2.50,  2.50,     0.00,     0.00);
           Tf1_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           --------------------------------------------------------------------
           -- V_fixture and R_fixture values
           --------------------------------------------------------------------
           Vfx_r1 : real := 0.0;
           Vfx_f1 : real := 0.0;

           Rfx_r1 : real := 50.0;
           Rfx_f1 : real := 50.0;
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_pu        : real    := 1.0;     -- PU current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling

           kt_rise      : real    := 1.0;     -- Rising waveform time scaling
           kt_fall      : real    := 1.0;     -- Falling waveform time scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           Max_dt       : real    := 1.0e-12; -- Maximum dt between time points
           Vth_R        : real    := 0.8;     -- Input threshold for rising edges
           Vth_F        : real    := 0.2;     -- Input threshold for falling edges
           MaxIV_length : integer := 100;     -- Maximum possible length of IV tables
           MaxVt_length : integer := 1000);   -- Maximum possible length of Vt tables
     --------------------------------------------------------------------------
     port (terminal PU_ref : electrical;
           terminal PD_ref : electrical;

           terminal Pad    : electrical;
           terminal In_D   : electrical;
           terminal En_D   : electrical;
--           terminal Rcv_D  : electrical;

           terminal PC_ref : electrical;
           terminal GC_ref : electrical);

end entity IBIS_IO_OPENSOURCE;
--=============================================================================
architecture IBIS_1EQ1UK of IBIS_IO_OPENSOURCE is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vpu   across  Ipu   through  PU_ref  to  Pad;
  quantity  Vpd   across  Ipd   through  Pad     to  PD_ref;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;

  quantity  Vin   across                 In_D    to  ELECTRICAL_REF;
  quantity  Ven   across                 En_D    to  ELECTRICAL_REF;
--  quantity  Vrcv  across  Ircv  through  Rcv_D   to  ELECTRICAL_REF;

  signal    pu_on         : integer := 0;
  signal    pu_off        : integer := 0;
  signal    EventState    : integer := 0;

  quantity  kpu           : current := 0.0;
  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_pu_val : real := FileRead(DataFile, "kC_comp_pu", kC_comp_pu);
  constant  kC_comp_pd_val : real := FileRead(DataFile, "kC_comp_pd", kC_comp_pd);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vinh_val    : real := FileRead(DataFile, "Vinh", Vinh);
  constant  Vinl_val    : real := FileRead(DataFile, "Vinl", Vinl);

  constant  Vpc_ref_val : real := FileRead(DataFile, "Vpc_ref", Vpc_ref);
  constant  Vpu_ref_val : real := FileRead(DataFile, "Vpu_ref", Vpu_ref);
  constant  Vgc_ref_val : real := FileRead(DataFile, "Vgc_ref", Vgc_ref);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Ipu_vec     : real_vector := FileRead(DataFile, "Ipu_data", Ipu_data, MaxIV_length);
  constant  Vpu_vec     : real_vector := FileRead(DataFile, "Vpu_data", Vpu_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);

  constant  Vr1_vec     : real_vector := FileRead(DataFile, "Vr1_data", Vr1_data, MaxVt_length);
  constant  Tr1_vec     : real_vector := FileRead(DataFile, "Tr1_data", Tr1_data, MaxVt_length, kt_rise);
  constant  Vf1_vec     : real_vector := FileRead(DataFile, "Vf1_data", Vf1_data, MaxVt_length);
  constant  Tf1_vec     : real_vector := FileRead(DataFile, "Tf1_data", Tf1_data, MaxVt_length, kt_fall);

  constant  Vfx_r1_val  : real := FileRead(DataFile, "Vfx_r1", Vfx_r1);
  constant  Vfx_f1_val  : real := FileRead(DataFile, "Vfx_f1", Vfx_f1);

  constant  Rfx_r1_val  : real := FileRead(DataFile, "Rfx_r1", Rfx_r1);
  constant  Rfx_f1_val  : real := FileRead(DataFile, "Rfx_f1", Rfx_f1);
--=============================================================================
  constant CommonLength : integer := FindCommonLength(Max_dt, Tr1_vec, Tf1_vec);

  type k_elements    is (Tcm, Kr1, Kf1);
  type k_tables_type is array(k_elements) of real_vector(1 to CommonLength);
--=============================================================================
  function Coeff (Common_length : integer;
                  Max_dt : real;
                  I_pc   : real_vector;
                  V_pc   : real_vector;
                  I_pu   : real_vector;
                  V_pu   : real_vector;
                  I_gc   : real_vector;
                  V_gc   : real_vector;

                  V_r1   : real_vector;
                  T_r1   : real_vector;
                  V_f1   : real_vector;
                  T_f1   : real_vector;

                  V_pc_ref : real;
                  V_pu_ref : real;
                  V_gc_ref : real;

                  V_fx_r1  : real;
                  V_fx_f1  : real;

                  R_fx_r1  : real;
                  R_fx_f1  : real;

                  C_comp_total : real) return k_tables_type is
  -----------------------------------------------------------------------------
  -- This function converts each Vt table to a corresponding scaling
  -- coefficient table that is used to scale the IV curves with respect
  -- to time.  The C_comp effects in the waveforms are eliminated from the
  -- scaling coefficients, so that the output waveforms will match the Vt
  -- table regardless of the presence of C_comp under identical loading
  -- conditions, i.e. C_comp in the output stage will not derate the
  -- original Vt curves.
  -----------------------------------------------------------------------------
    variable K_tables : k_tables_type;

    variable Vr1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf1cm : real_vector(1 to 3) := (others => 0.0);

    variable dVwfm_r1 : real := 0.0;
    variable dVwfm_f1 : real := 0.0;

    variable Iout1 : real := 0.0;
    variable Ipc1  : real := 0.0;
    variable Ipu1  : real := 0.0;
    variable Ipd1  : real := 0.0;
    variable Igc1  : real := 0.0;
  -----------------------------------------------------------------------------
  begin
    K_tables(Tcm) := CommonTime(Common_length, Max_dt, T_r1, T_f1);
    --=========================================================================
    -- Calculate the scaling coefficients for each time point along Tcm
    ---------------------------------------------------------------------------
    -- Store the first voltage point in the common waveform variables
    ---------------------------------------------------------------------------
    Vr1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r1, V_r1, "HE");
    Vf1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f1, V_f1, "HE");

    for i in K_tables(Tcm)'range loop
    --report "Tcm: " & real'image(K_tables(Tcm)(i));  -- For debugging
    ---------------------------------------------------------------------------
    -- Store the next point (for the derivative calculations)
    ---------------------------------------------------------------------------
      if (i < K_tables(Tcm)'right) then
        Vr1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r1, V_r1, "HE");
        Vf1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f1, V_f1, "HE");
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: " & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " R1: " & real'image(Vr1cm(2)) &
      --       " F1: " & real'image(Vf1cm(2));
      -------------------------------------------------------------------------
      -- Calculate the derivative of each waveform for C_comp compensation
      -------------------------------------------------------------------------
      if ((i <= K_tables(Tcm)'left) or (i >= K_tables(Tcm)'right)) then
        dVwfm_r1 := 0.0;          -- Force the end point derivatives to zero
        dVwfm_f1 := 0.0;
      else
        dVwfm_r1 := ((Vr1cm(2) - Vr1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr1cm(3) - Vr1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f1 := ((Vf1cm(2) - Vf1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf1cm(3) - Vf1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " dR1: " & real'image(dVwfm_r1) &
      --       " dF1: " & real'image(dVwfm_f1);
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kr1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vr1cm(2) - V_fx_r1) / R_fx_r1) + C_comp_total * dVwfm_r1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vr1cm(2), V_pc, I_pc, "HE");
      Ipu1  :=        Lookup(V_pu_ref - Vr1cm(2), V_pu, I_pu, "HE");
      Igc1  :=        Lookup(Vr1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are rising waveforms
      -- Kr1 == Kpu_on
      if (Ipu1 = 0.0) then K_tables(Kr1)(i) := 0.0;
      else                 K_tables(Kr1)(i) := (Ipc1 - Igc1 - Iout1) / Ipu1;
      end if;
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kf1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vf1cm(2) - V_fx_f1) / R_fx_f1) + C_comp_total * dVwfm_f1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vf1cm(2), V_pc, I_pc, "HE");
      Ipu1  :=        Lookup(V_pu_ref - Vf1cm(2), V_pu, I_pu, "HE");
      Igc1  :=        Lookup(Vf1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are falling waveforms
      -- Kf1 == Kpu_off
      if (Ipu1 = 0.0) then K_tables(Kf1)(i) := 0.0;
      else                 K_tables(Kf1)(i) := (Ipc1 - Igc1 - Iout1) / Ipu1;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " Kr1: " & real'image(K_tables(Kr1)(i)) &
      --       " Kf1: " & real'image(K_tables(Kf1)(i));
      -------------------------------------------------------------------------
      -- Shift points by one index for next iteration (for C_comp compensation)
      -------------------------------------------------------------------------
      Vr1cm(1) := Vr1cm(2);
      Vf1cm(1) := Vf1cm(2);

      Vr1cm(2) := Vr1cm(3);
      Vf1cm(2) := Vf1cm(3);

    end loop; -- for loop
    ---------------------------------------------------------------------------
    return K_tables;
  end function Coeff;
--=============================================================================
  constant Ktables : k_tables_type := Coeff(CommonLength, Max_dt,

                                            Ipc_vec, Vpc_vec,
                                            Ipu_vec, Vpu_vec,
                                            Igc_vec, Vgc_vec,

                                            Vr1_vec, Tr1_vec,
                                            Vf1_vec, Tf1_vec,

                                            Vpc_ref_val,
                                            Vpu_ref_val,
                                            Vgc_ref_val,

                                            Vfx_r1_val,
                                            Vfx_f1_val,

                                            Rfx_r1_val,
                                            Rfx_f1_val,

                                            C_comp_val);
--=============================================================================
begin

  Catch: process is
  -----------------------------------------------------------------------------
  begin
  -----------------------------------------------------------------------------
    if domain = quiescent_domain then

      wait until domain /= quiescent_domain;
      EventState <= 1;

    else

      wait on Vin'above(Vth_R), Vin'above(Vth_F),
              Ven'above(Vth_R), Ven'above(Vth_F);

      if ((Vin > Vth_R) or (Vin < Vth_F)) and
         ((Ven > Vth_R) or (Ven < Vth_F)) then

        if   EventState < 3 then  EventState <= EventState+1;
        else                      EventState <= EventState-1;
        end if;

      end if;
    end if;
  -----------------------------------------------------------------------------
  end process Catch;
  --===========================================================================
  FindState: process (EventState) is
  -----------------------------------------------------------------------------
  begin
  -----------------------------------------------------------------------------
    if (Ven > Vth_R) and (Vin > Vth_R) then           -- high state
      pu_on  <= 1;
      pu_off <= 0;
    elsif (Ven > Vth_R) and (Vin < Vth_F) then        -- low state
      pu_on  <= 0;
      pu_off <= 1;
    elsif (Ven < Vth_F) then                          -- 3-state
      pu_on  <= 0;
      pu_off <= 1;
    else                                              -- unknown state
      pu_on  <= 1;
      pu_off <= 0;
    end if;
  -----------------------------------------------------------------------------
  end process FindState;
--=============================================================================
  break on pu_on;
  break on pu_off;
--=============================================================================
  -- This section contains the simultaneous analog equations to find the
  -- appropriate scaling coefficients according to the state the buffer.
  -----------------------------------------------------------------------------
--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
-- This entire section could be made more efficient with nested IF statements
-- if the simulator wouldn't have a problem with simultaneous statements in
-- nested case statements.
--XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

  -- Executes during DC operating point calculation
  if    (EventState = 0) and (Ven > Vth_R) and (Vin > Vth_R) use   -- high state
    kpu == Ktables(Kr1)(CommonLength);
  elsif (EventState = 0) and (Ven > Vth_R) and (Vin < Vth_F) use   -- low state
    kpu == Ktables(Kf1)(CommonLength);
  elsif (EventState = 0) and (Ven < Vth_F) use                     -- 3-state
    kpu == Ktables(Kf1)(CommonLength);

  -- Executes after DC operating point calculation before first event
  elsif (EventState = 1) and (pu_off = 1) use
    kpu == Ktables(Kf1)(CommonLength);      -- Last point from pu_off table
  elsif (EventState = 1) and (pu_off /= 1)use
    kpu == Ktables(Kr1)(CommonLength);      -- Last point from pu_on table

  -- Executes after first event (normal operation)
  elsif (EventState > 1) and (pu_on  = 1) use    -- Pullup turning on
    kpu == Lookup(pu_on'last_event,  Ktables(Tcm), Ktables(Kr1), "HE");
  elsif (EventState > 1) and (pu_off = 1) use    -- Pullup turning off
    kpu == Lookup(pu_off'last_event, Ktables(Tcm), Ktables(Kf1), "HE");

  -- Just in case
  else
    kpu == Ktables(Kr1)(Ktables(Kr1)'left); -- 1st point of pullup on table
  end use;

--=============================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc == -1.0*kI_pc    *Lookup(Vpc, Vpc_vec, Ipc_vec, "SE");-- + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Ipu == -1.0*kI_pu*kpu*Lookup(Vpu, Vpu_vec, Ipu_vec, "SE");-- + kC_comp_pu_val*C_comp_val*Vpu'dot;
  Ipd ==  0.0                                              ;-- + kC_comp_pd_val*C_comp_val*Vpd'dot;
  Igc ==      kI_gc    *Lookup(Vgc, Vgc_vec, Igc_vec, "SE");-- + kC_comp_gc_val*C_comp_val*Vgc'dot;
  -----------------------------------------------------------------------------
  -- A simple receiver logic
  -----------------------------------------------------------------------------
--  if    (Vgc > Vinh_val) use Vrcv == 1.0;
--  elsif (Vgc < Vinl_val) use Vrcv == 0.0;
--  else                       Vrcv == 0.5;
--  end use;
--=============================================================================
end architecture IBIS_1EQ1UK;
--=============================================================================
